from typing import Callable, Dict

import gym
import numpy as np
import numpy.typing as npt
from gym import spaces

import luxai_s2.env
from luxai_s2.env import LuxAI_S2
from luxai_s2.state import ObservationStateDict
from luxai_s2.unit import ActionType, BidActionType, FactoryPlacementActionType
from luxai_s2.utils import my_turn_to_place_factory
from luxai_s2.wrappers.controllers import (
    Controller,
    SimpleDiscreteController,
    SimpleSingleUnitDiscreteController,
)


class SB3Wrapper(gym.Wrapper):
    def __init__(
        self,
        env: LuxAI_S2,
        bid_policy: Callable[
            [str, ObservationStateDict], Dict[str, BidActionType]
        ] = None,
        factory_placement_policy: Callable[
            [str, ObservationStateDict], Dict[str, FactoryPlacementActionType]
        ] = None,
        heuristic_policy: Callable[
            [str, ObservationStateDict], Dict[str, ActionType]
        ] = None,
        controller: Controller = None,
    ) -> None:
        """
        A environment wrapper for Stable Baselines 3. It reduces the LuxAI_S2 env
        into a single phase game and places the first two phases (bidding and factory placement) into the env.reset function so that
        interacting agents directly start generating actions to play the third phase of the game.

        It's highly recommended to use one of the observation wrappers as well

        Parameters
        ----------
        bid_policy: Function
            A function accepting player: str and obs: ObservationStateDict as input that returns a bid action
            such as dict(bid=10, faction="AlphaStrike"). By default will bid 0
        factory_placement_policy: Function
            A function accepting player: str and obs: ObservationStateDict as input that returns a factory placement action
            such as dict(spawn=np.array([2, 4]), metal=150, water=150). By default will spawn in a random valid location with metal=150, water=150
        controller : Controller
            A controller that parameterizes the action space into something more usable and converts parameterized actions to lux actions.
            See luxai_s2/wrappers/controllers.py for available controllers and how to make your own

        heuristic_policy: Function
            A function accepting player: str and obs: ObservationStateDict as input and returns a lux action. This can be provided by the user
            to define custom logic or a model to generate actions for any of the units or factories. For any action generate for a unit or factory, it will
            override the original action for that unit or factory when the step function is called. By defalt this is None and not used
        """
        gym.Wrapper.__init__(self, env)
        self.env = env
        if controller is None:
            controller = SimpleDiscreteController(self.env.state.env_cfg)
        self.controller = controller

        self.action_space = controller.action_space

        obs_dims = 23  # see _convert_obs function for how this is computed
        self.map_size = self.env.env_cfg.map_size
        self.observation_space = spaces.Box(
            -999, 999, shape=(self.map_size, self.map_size, obs_dims)
        )

        # The simplified wrapper removes the first two phases of the game by using predefined policies (trained or heuristic)
        # to handle those two phases during each reset
        if factory_placement_policy is None:

            def factory_placement_policy(player, obs: ObservationStateDict):
                potential_spawns = np.array(
                    list(zip(*np.where(obs["board"]["valid_spawns_mask"] == 1)))
                )
                spawn_loc = potential_spawns[
                    np.random.randint(0, len(potential_spawns))
                ]
                return dict(spawn=spawn_loc, metal=150, water=150)

        self.factory_placement_policy = factory_placement_policy
        if bid_policy is None:

            def bid_policy(player, obs: ObservationStateDict):
                faction = "AlphaStrike"
                if player == "player_1":
                    faction = "MotherMars"
                return dict(bid=0, faction=faction)

        self.bid_policy = bid_policy

        self.heuristic_policy = heuristic_policy

        self.prev_obs = None
        # list of all agents regardless of status
        self.all_agents = []

    def step(self, action: Dict[str, npt.NDArray]):
        lux_action = dict()
        for agent in self.all_agents:
            if agent in action:
                lux_action[agent] = self.controller.action_to_lux_action(
                    agent=agent, obs=self.prev_obs, action=action[agent]
                )
            else:
                lux_action[agent] = dict()
            if self.heuristic_policy is not None:
                heuristic_lux_action = self.heuristic_policy(
                    agent, self.prev_obs[agent]
                )
                # override keys
                for k in heuristic_lux_action:
                    lux_action[agent][k] = heuristic_lux_action[k]
        obs, reward, done, info = self.env.step(lux_action)
        self.prev_obs = obs
        return obs, reward, done, info

    def reset(self, **kwargs):
        obs = self.env.reset(**kwargs)
        self.all_agents = self.env.agents
        action = dict()
        for agent in self.all_agents:
            action[agent] = self.bid_policy(agent, obs[agent])
        obs, _, _, _ = self.env.step(action)
        while self.env.state.real_env_steps < 0:
            action = dict()
            for agent in self.all_agents:
                if my_turn_to_place_factory(
                    obs["player_0"]["teams"][agent]["place_first"],
                    self.env.state.env_steps,
                ):
                    action[agent] = self.factory_placement_policy(agent, obs[agent])
                else:
                    action[agent] = dict()
            obs, _, _, _ = self.env.step(action)
        self.prev_obs = obs
        return obs



import asyncio
import json
import sys
from pathlib import Path
from typing import Dict, List

import numpy as np
from luxai_runner.bot import Bot
from luxai_runner.episode import Episode, EpisodeConfig, ReplayConfig
from luxai_runner.logger import Logger
from luxai_runner.tournament import Tournament, TournamentConfig
from omegaconf import OmegaConf

from luxai_s2 import LuxAI_S2


def main():
    np.random.seed(0)
    import argparse

    parser = argparse.ArgumentParser(description="Run the LuxAI Season 2 game.")
    parser.add_argument(
        "players",
        nargs="+",
        help="Paths to player modules. If --tournament is passed as well, you can also pass a folder and we will look through all sub-folders for valid agents with main.py files (only works for python agents at the moment).",
    )
    parser.add_argument(
        "-l", "--len", help="Max episode length", type=int, default=1000
    )

    # replay configs
    parser.add_argument(
        "-o",
        "--output",
        help="Where to output replays. Default is none and no replay is generated",
    )
    parser.add_argument(
        "--replay.save_format",
        help='Format of the replay file, can be either "html" or "json". HTML replays are easier to visualize, JSON replays are easier to analyze programmatically. Defaults to the extension of the path passed to --output, or "json" if there is no extension or it is invalid.',
        default="json",
    )
    parser.add_argument(
        "--replay.compressed_obs",
        help="Whether to save compressed observations or not. Compressed observations do not contain the full observation at each step. In particular, the map information is stored as the first observation, subsequent observations only store the changes that happened.",
        default=True,
    )

    # episode configs
    parser.add_argument(
        "-v",
        "--verbose",
        help="Verbose Level (0 = silent, 1 = errors, 2 = warnings, 3 = info)",
        type=int,
        default=1,
    )
    parser.add_argument(
        "-s",
        "--seed",
        help="Fix a seed for episode(s). All episodes will initialize the same, including tournament ones",
        type=int,
    )

    # env configs

    parser.add_argument(
        "--render", help="Render with a window", action="store_true", default=False
    )

    parser.add_argument(
        "--tournament",
        help="Turn tournament mode on",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--tournament_cfg.concurrent",
        help="Max concurrent number of episodes to run. Recommended to set no higher than the number of CPUs / 2",
        type=int,
        default=1,
    )
    parser.add_argument(
        "--tournament_cfg.ranking_system",
        help="The ranking system to use. Default is 'elo'. Can be 'elo', 'wins'.",
        type=str,
        default="elo",
    )
    parser.add_argument(
        "--skip_validate_action_space",
        help="Set this for a small performance increase. Note that turning this on means the engine assumes your submitted actions are valid. If your actions are not well formatted there could be errors",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    save_format = getattr(args, "replay.save_format")
    if args.output is not None:
        output_file = Path(args.output).name
        if "." in output_file:
            output_ext = args.output.split(".")[-1]
            if output_ext in ["html", "json"]:
                save_format = output_ext

    cfg = EpisodeConfig(
        players=args.players,
        env_cls=LuxAI_S2,
        seed=args.seed,
        env_cfg=dict(
            verbose=args.verbose,
            validate_action_space=not args.skip_validate_action_space,
            max_episode_length=args.len,
        ),
        verbosity=args.verbose,
        save_replay_path=args.output,
        replay_options=ReplayConfig(
            save_format=save_format,
            compressed_obs=getattr(args, "replay.compressed_obs"),
        ),
        render=args.render,
    )

    if args.tournament:
        import os

        if os.path.isdir(args.players[0]):
            assert (
                len(args.players) == 1
            ), "Found more than one positional argument despite being given a directory of players"
            subfolders = [f.path for f in os.scandir(args.players[0]) if f.is_dir()]
            agents = []
            for sub_dir in subfolders:
                agent_file = os.path.join(sub_dir, "main.py")
                if os.path.isfile(agent_file):
                    agents.append(agent_file)
            print(f"Found {len(agents)} in {args.players[0]}")
            args.players = agents

        tournament_config = TournamentConfig()
        tournament_config.agents = args.players
        # TODO - in future replace this with OmegaConf or something that can parse these nicely
        tournament_config.max_concurrent_episodes = getattr(
            args, "tournament_cfg.concurrent"
        )
        tournament_config.ranking_system = getattr(
            args, "tournament_cfg.ranking_system"
        )
        tourney = Tournament(
            cfg=tournament_config, episode_cfg=cfg  # the base/default episode config
        )
        # import ipdb;ipdb.set_trace()
        asyncio.run(tourney.run())
        # exit()
    else:
        import time

        stime = time.time()
        eps = Episode(cfg=cfg)
        asyncio.run(eps.run())
        etime = time.time()
        print(etime - stime)


if __name__ == "__main__":
    main()



import asyncio
import copy
import os
import os.path as osp
from typing import Dict

from luxai_runner.episode import Episode, EpisodeConfig
from luxai_runner.tournament import matchmaking
from luxai_runner.tournament.config import TournamentConfig
from luxai_runner.tournament.rankingsystem import ELO, Rank, WinLoss, WinLossRank


class Player:
    def __init__(self, player_id, file) -> None:
        self.id = player_id
        self.file = file
        self.rank: Rank = None


class Tournament:
    def __init__(self, cfg: TournamentConfig, episode_cfg: EpisodeConfig = None):
        self.global_id = 0
        self.episode_id = 0
        self.cfg = cfg
        self.eps_cfg = episode_cfg
        min_agents = min(self.cfg.agents_per_episode)
        assert len(self.cfg.agents) >= min_agents
        assert episode_cfg is not None

        # init ranking system
        if cfg.ranking_system.lower() == "elo":
            # TODO - allow configuration via CLI
            self.ranking_sys = ELO(K=30, init_rating=1000)
        elif cfg.ranking_system.lower() == "wins":
            self.ranking_sys = WinLoss(win_points=3, tie_points=1, loss_points=0)
        else:
            raise Exception(f"'{cfg.ranking_system}' is not a valid ranking system")

        self.players: Dict[str, Player] = dict()
        for agent in self.cfg.agents:
            self.add_player(agent)

        self.match_making_sys = matchmaking.Random(
            [x for x in self.players], self.cfg.agents_per_episode
        )

    def add_player(self, file, name=None):
        if name is None:
            name = file
        name = f"{name}_{self.global_id}"
        assert os.path.isfile(file), f"Agent file {file} does not exist"

        self.global_id += 1
        self.players[name] = Player(name, file)
        self.players[name].rank = self.ranking_sys.init_rank_state()

    async def run(self):
        episodes = set()

        async def _run_episode_cb(a):
            if a in episodes:
                episodes.discard(a)
            next_players = self.match_making_sys.next_match()
            eps_cfg = copy.deepcopy(self.eps_cfg)
            if self.eps_cfg.save_replay_path is not None:
                # if there is a replay path template, use it and generate a replay path for this episode
                save_replay_path_split = osp.splitext(eps_cfg.save_replay_path)
                eps_cfg.save_replay_path = f"{save_replay_path_split[0]}_{self.episode_id}{save_replay_path_split[1]}"
            eps_cfg.players = [self.players[p].file for p in next_players]
            self.episode_id += 1

            players = dict()
            for i in range(len(next_players)):
                players[f"player_{i}"] = self.players[next_players[i]]

            task = asyncio.Task(self._run_episode(players, eps_cfg))
            episodes.add(task)
            await task
            await _run_episode_cb(task)

        async def print_results():
            import time

            line_length = 50
            line_length += len(self.ranking_sys._rank_headers())
            while True:
                import sys
                import time
                from collections import deque

                lines = []

                lines.append(f"==== {self.cfg.name} ====")
                # lines.append("")
                lines.append(
                    f"{'Player':36.36}| {self.ranking_sys._rank_headers()}| {'Episodes':14.14}"
                )
                lines.append("-" * line_length)
                players_sorted = [self.players[p] for p in self.players]
                players_sorted = sorted(
                    players_sorted, key=lambda p: p.rank.rating, reverse=True
                )
                for p in players_sorted:
                    rank = p.rank
                    lines.append(
                        f"{p.id:36.36}| {self.ranking_sys._rank_info(rank)}| {str(rank.episodes):14.14}"
                    )
                lines.append("-" * line_length)
                lines.append(f"{len(episodes)} episodes are running")

                for _ in range(len(lines)):
                    sys.stdout.write("\x1b[1A\x1b[2K")
                for i in range(len(lines)):
                    sys.stdout.write(lines[i] + "\n")
                await asyncio.sleep(1)

        coros = []
        for i in range(self.cfg.max_concurrent_episodes):
            coros += [_run_episode_cb(None)]
        await asyncio.gather(*coros, print_results())

    async def _run_episode(self, players: Dict[str, Player], eps_cfg: EpisodeConfig):
        eps = Episode(cfg=eps_cfg)
        rewards = await eps.run()
        # TODO: there is probably some race condition here, but in long run it's not important
        self.ranking_sys.update(
            players["player_0"].rank,
            players["player_1"].rank,
            rewards["player_0"],
            rewards["player_1"],
        )



from __future__ import annotations

import time
from collections import deque
from itertools import product
from typing import List
try:
    from typing import TypedDict    
except:
    from typing_extensions import TypedDict

import numpy as np
import numpy.typing as npt

from luxai_s2.actions import move_deltas
from luxai_s2.config import EnvConfig
from luxai_s2.globals import TERM_COLORS
from luxai_s2.map.board import Board
from luxai_s2.map.position import Position
from luxai_s2.team import Team
from luxai_s2.unit import UnitCargo, UnitCargoStateDict

try:
    from termcolor import colored
except:
    pass


def compute_water_info(
    init: np.ndarray,
    MIN_LICHEN_TO_SPREAD: int,
    lichen: np.ndarray,
    lichen_strains: np.ndarray,
    factory_occupancy_map: np.ndarray,
    strain_id: int,
    forbidden: np.ndarray,
):
    # TODO - improve the performance here with cached solution
    frontier = deque(init)
    seen = set(map(tuple, init))
    grow_lichen_positions = set()
    H, W = lichen.shape
    ct = 0
    while len(frontier) > 0:
        ct += 1
        if ct > 1_000_000:
            print("Error! Lichen Growth calculation took too long")
            break
        pos = frontier.popleft()
        if (
            pos[0] < 0
            or pos[1] < 0
            or pos[0] >= forbidden.shape[0]
            or pos[1] >= forbidden.shape[1]
        ):
            continue

        if forbidden[pos[0], pos[1]]:
            continue
        pos_lichen = lichen[pos[0], pos[1]]
        pos_strain = lichen_strains[pos[0], pos[1]]
        # check for surrounding tiles with lichen and no incompatible lichen strains, grow on those
        can_grow = True
        for move_delta in move_deltas[1:]:
            check_pos = pos + move_delta
            # check surrounding tiles on the map
            if (
                check_pos[0] < 0
                or check_pos[1] < 0
                or check_pos[0] >= H
                or check_pos[1] >= W
            ):
                continue

            # If any neighbor 1. has a different strain, or 2. is a different factory,
            # then the current pos cannot grow
            adj_strain = lichen_strains[check_pos[0], check_pos[1]]
            adj_factory = factory_occupancy_map[check_pos[0], check_pos[1]]
            if (adj_strain != -1 and adj_strain != strain_id) or (
                adj_factory != -1 and adj_factory != strain_id
            ):
                can_grow = False

            # if seen, skip
            if (check_pos[0], check_pos[1]) in seen:
                continue

            # we add it to the frontier only in two cases:
            #  1. it is an empty tile, and current pos has enough lichen to expand.
            #  2. both current tile and check_pos are of our strain.
            if (adj_strain == -1 and pos_lichen >= MIN_LICHEN_TO_SPREAD) or (
                adj_strain == strain_id and pos_strain == strain_id
            ):
                seen.add(tuple(check_pos))
                frontier.append(check_pos)

        if can_grow or (lichen_strains[pos[0], pos[1]] == strain_id):
            grow_lichen_positions.add((pos[0], pos[1]))
    return grow_lichen_positions


class FactoryStateDict(TypedDict):
    pos: npt.NDArray[np.int_]
    power: int
    cargo: UnitCargoStateDict
    unit_id: str
    strain_id: int
    team_id: int


class Factory:
    def __init__(self, team: Team, unit_id: str, num_id: int) -> None:
        self.team_id = team.team_id
        self.team = team
        self.unit_id = unit_id
        self.pos = Position(np.zeros(2, dtype=int))
        self.power = 0
        self.cargo = UnitCargo()
        self.num_id = num_id
        self.action_queue = []
        self.grow_lichen_positions = set()

    @property
    def pos_slice(self):
        return slice(self.pos.x - 1, self.pos.x + 2), slice(
            self.pos.y - 1, self.pos.y + 2
        )

    @property
    def min_dist_slice(self):
        deltas = np.array(
            [
                (4, 0),
                (5, 1),
                (0, -4),
                (0, 5),
                (2, 2),
                (-4, 1),
                (-1, -1),
                (-1, -2),
                (-2, -1),
                (-2, -2),
                (4, 2),
                (0, -2),
                (0, -1),
                (2, 4),
                (-1, 0),
                (-2, 0),
                (-3, 1),
                (0, 0),
                (2, -3),
                (-2, 2),
                (-1, 2),
                (3, 1),
                (5, -1),
                (-3, 3),
                (0, 2),
                (1, 3),
                (-4, -1),
                (-1, -5),
                (-4, -2),
                (-1, 4),
                (-2, 4),
                (3, 3),
                (5, 0),
                (-6, 0),
                (0, -5),
                (1, -4),
                (1, 5),
                (-4, 0),
                (-1, -3),
                (-2, -3),
                (-5, 1),
                (-3, -1),
                (0, -3),
                (-3, -2),
                (1, -1),
                (1, -2),
                (-4, 2),
                (3, -1),
                (3, -2),
                (-3, 0),
                (1, 0),
                (3, 0),
                (-3, 2),
                (1, 2),
                (0, 4),
                (2, 1),
                (3, 2),
                (4, 1),
                (-5, -1),
                (1, -5),
                (1, 4),
                (0, 6),
                (2, 3),
                (-1, -4),
                (-5, 0),
                (-3, -3),
                (1, -3),
                (2, -4),
                (-1, 1),
                (-2, 1),
                (3, -3),
                (0, 1),
                (2, -1),
                (2, -2),
                (-1, 3),
                (-2, 3),
                (4, -1),
                (4, -2),
                (0, -6),
                (1, 1),
                (0, 3),
                (2, 0),
                (6, 0),
                (-2, -4),
                (-1, 5),
            ]
        )
        return self.pos.pos + deltas

    def refine_step(self, config: EnvConfig):
        max_consumed_ice = min(self.cargo.ice, config.FACTORY_PROCESSING_RATE_WATER)
        max_consumed_ore = min(self.cargo.ore, config.FACTORY_PROCESSING_RATE_METAL)
        # permit refinement of blocks of resources, no floats.
        produced_water = max_consumed_ice // config.ICE_WATER_RATIO
        produced_metal = max_consumed_ore // config.ORE_METAL_RATIO
        self.cargo.ice -= produced_water * config.ICE_WATER_RATIO
        self.cargo.ore -= produced_metal * config.ORE_METAL_RATIO

        self.cargo.water += produced_water
        self.cargo.metal += produced_metal

    def cache_water_info(self, board: Board, env_cfg: EnvConfig):
        # TODO this can easily be a fairly slow function, can we make it much faster?
        # Caches information about which tiles lichen can grow on for this factory

        # perform a BFS from the factory position and look for non rubble, non factory tiles.
        # find the current frontier from 12 starting positions x marked below
        """
           x x x
           _ _ _
        x |     | x
        x |     | x
        x |_ _ _| x
           x x x

        """
        forbidden = (
            (board.rubble > 0)
            | (board.factory_occupancy_map != -1)
            | (board.ice > 0)
            | (board.ore > 0)
        )
        deltas = [
            np.array([0, -2]),
            np.array([-1, -2]),
            np.array([1, -2]),
            np.array([0, 2]),
            np.array([-1, 2]),
            np.array([1, 2]),
            np.array([2, 0]),
            np.array([2, -1]),
            np.array([2, 1]),
            np.array([-2, 0]),
            np.array([-2, -1]),
            np.array([-2, 1]),
        ]
        init_arr = np.stack(deltas) + self.pos.pos
        self.grow_lichen_positions = compute_water_info(
            init_arr,
            env_cfg.MIN_LICHEN_TO_SPREAD,
            board.lichen,
            board.lichen_strains,
            board.factory_occupancy_map,
            self.num_id,
            forbidden,
        )

    def water_cost(self, config: EnvConfig):
        return int(
            np.ceil(
                len(self.grow_lichen_positions) / config.LICHEN_WATERING_COST_FACTOR
            )
        )

    ### Add and sub resource functions copied over from unit.py code, can we consolidate them somewhere?
    def add_resource(self, resource_id, transfer_amount):
        if transfer_amount < 0:
            transfer_amount = 0
        if resource_id == 0:
            self.cargo.ice += transfer_amount
        elif resource_id == 1:
            self.cargo.ore += transfer_amount
        elif resource_id == 2:
            self.cargo.water += transfer_amount
        elif resource_id == 3:
            self.cargo.metal += transfer_amount
        elif resource_id == 4:
            self.power += transfer_amount
        return int(transfer_amount)

    def sub_resource(self, resource_id, amount):
        # subtract/transfer out as much as you min(have, request)
        if amount < 0:
            amount = 0
        if resource_id == 0:
            transfer_amount = min(self.cargo.ice, amount)
            self.cargo.ice -= transfer_amount
        elif resource_id == 1:
            transfer_amount = min(self.cargo.ore, amount)
            self.cargo.ore -= transfer_amount
        elif resource_id == 2:
            transfer_amount = min(self.cargo.water, amount)
            self.cargo.water -= transfer_amount
        elif resource_id == 3:
            transfer_amount = min(self.cargo.metal, amount)
            self.cargo.metal -= transfer_amount
        elif resource_id == 4:
            transfer_amount = min(self.power, amount)
            self.power -= transfer_amount
        return int(transfer_amount)

    def state_dict(self) -> FactoryStateDict:
        return dict(
            pos=self.pos.pos,
            power=self.power,
            cargo=self.cargo.state_dict(),
            unit_id=self.unit_id,
            strain_id=self.num_id,  # number version of unit_id
            team_id=self.team_id,
        )

    def __str__(self) -> str:
        out = f"[{self.team_id}] {self.unit_id} Factory at {self.pos}"
        if TERM_COLORS:
            return colored(out, self.team.faction.value.color)
        return out



from abc import ABC
from typing import List

import numpy as np


class MatchMakingSystem(ABC):
    def __init__(
        self, agents: List[str], agents_per_episode: List[int], seed=0
    ) -> None:
        super().__init__()
        self.agents = agents
        self.agents_per_episode = agents_per_episode
        self.np_random = np.random.RandomState(seed=seed)

    def next_match(self) -> List[str]:
        pass


class Random(MatchMakingSystem):
    def __init__(
        self, agents: List[str], agents_per_episode: List[int], seed=0
    ) -> None:
        super().__init__(agents, agents_per_episode, seed)

    def next_match(self) -> List[str]:
        num_agents = self.np_random.choice(self.agents_per_episode)
        agents = self.np_random.choice(self.agents, size=num_agents, replace=False)
        return agents



import math
from dataclasses import dataclass
from enum import Enum
from typing import List
try:
    from typing import TypedDict    
except:
    from typing_extensions import TypedDict

import numpy as np
import numpy.typing as npt

try:
    from termcolor import colored
except:
    pass
from luxai_s2.config import EnvConfig, UnitConfig
from luxai_s2.globals import TERM_COLORS
from luxai_s2.map.position import Position
from luxai_s2.team import FactionTypes, Team


class UnitType(Enum):
    LIGHT = "Light"
    HEAVY = "Heavy"


class UnitCargoStateDict(TypedDict):
    ice: int
    ore: int
    water: int
    metal: int


ActionType = npt.NDArray[np.int_]


class FactoryPlacementActionType(TypedDict):
    metal: int
    water: int
    spawn: npt.NDArray[np.int_]


class BidActionType(TypedDict):
    faction: str
    bid: int


class UnitStateDict(TypedDict):
    team_id: int
    unit_id: str
    power: int
    unit_type: str
    pos: npt.NDArray[np.int_]
    cargo: UnitCargoStateDict
    action_queue: List[ActionType]


@dataclass
class UnitCargo:
    ice: int = 0
    ore: int = 0
    water: int = 0
    metal: int = 0

    def state_dict(self) -> UnitCargoStateDict:
        return dict(ice=self.ice, ore=self.ore, water=self.water, metal=self.metal)


class Unit:
    def __init__(
        self, team: Team, unit_type: UnitType, unit_id: str, env_cfg: EnvConfig
    ) -> None:
        self.unit_type = unit_type
        self.team_id = team.team_id
        self.team = team
        self.unit_id = unit_id
        self.pos = Position(np.zeros(2, dtype=int))

        self.cargo = UnitCargo()
        # TODO - replace with a deque perhaps?
        self.action_queue: List = []
        self.unit_cfg: UnitConfig = env_cfg.ROBOTS[unit_type.name]
        self.power = env_cfg.ROBOTS[unit_type.name].INIT_POWER
        self.cargo_space = env_cfg.ROBOTS[unit_type.name].CARGO_SPACE
        self.battery_capacity = env_cfg.ROBOTS[unit_type.name].BATTERY_CAPACITY

    def __str__(self) -> str:
        out = f"[{self.team_id}] {self.unit_id} {self.unit_type} at {self.pos}"
        if TERM_COLORS:
            return colored(out, self.team.faction.value.color)
        return out

    def is_heavy(self) -> bool:
        return self.unit_type == UnitType.HEAVY

    def next_action(self):
        """
        get next action
        """
        if len(self.action_queue) == 0:
            return None
        action = self.action_queue[0]
        return action

    def repeat_action(self, action):
        action.n -= 1
        if action.n <= 0:
            # remove from front of queue
            self.action_queue.pop(0)
            # endless repeat puts action back at end of queue
            if action.repeat:
                action.n = 1
                self.action_queue.append(action)

    def move_power_cost(self, rubble_at_target: int):
        return math.floor(
            self.unit_cfg.MOVE_COST
            + self.unit_cfg.RUBBLE_MOVEMENT_COST * rubble_at_target
        )

    def state_dict(self) -> UnitStateDict:
        return dict(
            team_id=self.team_id,
            unit_id=self.unit_id,
            power=self.power,
            unit_type=self.unit_type.name,
            pos=self.pos.pos,
            cargo=self.cargo.state_dict(),
            action_queue=[a.state_dict() for a in self.action_queue],
        )

    def add_resource(self, resource_id, amount):
        if amount < 0:
            amount = 0
        if resource_id == 0:
            transfer_amount = min(self.cargo_space - self.cargo.ice, amount)
            self.cargo.ice += transfer_amount
        elif resource_id == 1:
            transfer_amount = min(self.cargo_space - self.cargo.ore, amount)
            self.cargo.ore += transfer_amount
        elif resource_id == 2:
            transfer_amount = min(self.cargo_space - self.cargo.water, amount)
            self.cargo.water += transfer_amount
        elif resource_id == 3:
            transfer_amount = min(self.cargo_space - self.cargo.metal, amount)
            self.cargo.metal += transfer_amount
        elif resource_id == 4:
            transfer_amount = min(self.battery_capacity - self.power, amount)
            self.power += transfer_amount
        return int(transfer_amount)

    def sub_resource(self, resource_id, amount):
        # subtract/transfer out as much as you min(have, request)
        if amount < 0:
            amount = 0
        if resource_id == 0:
            transfer_amount = min(self.cargo.ice, amount)
            self.cargo.ice -= transfer_amount
        elif resource_id == 1:
            transfer_amount = min(self.cargo.ore, amount)
            self.cargo.ore -= transfer_amount
        elif resource_id == 2:
            transfer_amount = min(self.cargo.water, amount)
            self.cargo.water -= transfer_amount
        elif resource_id == 3:
            transfer_amount = min(self.cargo.metal, amount)
            self.cargo.metal -= transfer_amount
        elif resource_id == 4:
            transfer_amount = min(self.power, amount)
            self.power -= transfer_amount
        return int(transfer_amount)


if __name__ == "__main__":
    u = Unit(team=Team(0, FactionTypes.AlphaStrike), unit_type=UnitType.HEAVY)
    print(u)
    u = Unit(team=Team(1, FactionTypes.MotherMars), unit_type=UnitType.HEAVY)
    u.pos += np.array([1, 4])
    print(u)
    # u = Unit(team=Team(0, FactionTypes.AlphaStrike), unit_type=UnitType.HEAVY)
    # print(u)



import json
import os
import os.path as osp
from typing import List, Tuple

from luxai_s2.state import State


def decode_replay_file(replay_file) -> Tuple[State, List]:
    """
    Takes an input replay file of any kind from any source and extracts the full trajectory with observations
    or just initial state and actions if observations aren't there
    """
    ext = osp.splitext(replay_file)[-1]
    if ext == ".json":
        # probably kaggle replay
        with open(replay_file, "r") as f:
            replay = json.load(f)
        init_state = replay["init_state"]
        pass
    elif ext == ".h5":
        pass


def generate_replay(states: List[State]):
    """
    Generates a compressed replay.

    """
    return [s.get_compressed_obs() for s in states]



from typing import Dict

import numpy as np
import numpy.typing as npt
from gym import spaces

from luxai_s2.actions import format_action_vec
from luxai_s2.config import EnvConfig
from luxai_s2.state import ObservationStateDict


class Controller:
    def __init__(self, action_space: spaces.Space) -> None:
        self.action_space = action_space

    def action_to_lux_action(
        self, agent: str, obs: Dict[str, ObservationStateDict], action: npt.NDArray
    ):
        """
        Takes as input the current "raw observation" and the parameterized action and returns
        an action formatted for the Lux env
        """
        raise NotImplementedError()


class SimpleSingleUnitDiscreteController(Controller):
    def __init__(self, env_cfg: EnvConfig) -> None:
        """
        A simple controller that controls only the heavy unit that will get spawned. This assumes for whichever environment wrapper you use
        you have defined a policy to generate the first factory action

        For the heavy unit
        - 4 cardinal direction movement (4 dims)
        - a move center no-op action (1 dim)
        - transfer action just for transferring ice in 4 cardinal directions or center (5)
        - pickup action for each resource (5 dims)
        - dig action (1 dim)

        It does not include
        - self destruct action
        - recharge action
        - planning (via actions executing multiple times or repeating actions)
        - factory actions
        - transferring power or resources other than ice
        """
        self.env_cfg = env_cfg
        self.move_act_dims = 5
        self.transfer_act_dims = 5  # 5 * 5
        self.pickup_act_dims = 5
        self.dig_act_dims = 1

        self.move_dim_high = self.move_act_dims
        self.transfer_dim_high = self.move_dim_high + self.transfer_act_dims
        self.pickup_dim_high = self.transfer_dim_high + self.pickup_act_dims
        self.dig_dim_high = self.pickup_dim_high + self.dig_act_dims

        total_act_dims = self.dig_dim_high
        # action_space = spaces.Box(0, 1, shape=(total_act_dims,))
        action_space = spaces.Discrete(total_act_dims)
        super().__init__(action_space)

    def _is_move_action(self, id):
        return id < self.move_dim_high

    def _get_move_action(self, id):
        return np.array([0, id, 0, 0, 0, 1])

    def _is_transfer_action(self, id):
        return id < self.transfer_dim_high

    def _get_transfer_action(self, id):
        id = id - self.move_dim_high
        transfer_dir = id % 5
        # resource_type = id // 5
        return np.array([1, transfer_dir, 0, self.env_cfg.max_transfer_amount, 0, 1])

    def _is_pickup_action(self, id):
        return id < self.pickup_dim_high

    def _get_pickup_action(self, id):
        id = id - self.transfer_dim_high
        return np.array([2, 0, id % 5, self.env_cfg.max_transfer_amount, 0, 1])

    def _is_dig_action(self, id):
        return id < self.dig_dim_high

    def _get_dig_action(self, id):
        return np.array([3, 0, 0, 0, 0, 1])

    def action_to_lux_action(
        self, agent: str, obs: Dict[str, ObservationStateDict], action: npt.NDArray
    ):
        shared_obs = obs["player_0"]
        lux_action = dict()
        factories = shared_obs["factories"][agent]
        units = shared_obs["units"][agent]
        for unit_id in units.keys():
            unit = units[unit_id]
            pos = unit["pos"]
            unit_related_action = action
            choice = action  # unit_related_action.argmax()
            action_queue = []
            if self._is_move_action(choice):
                action_queue = [self._get_move_action(choice)]
            elif self._is_transfer_action(choice):
                action_queue = [self._get_transfer_action(choice)]
            elif self._is_pickup_action(choice):
                action_queue = [self._get_pickup_action(choice)]

            elif self._is_dig_action(choice):
                action_queue = [self._get_dig_action(choice)]
            lux_action[unit_id] = action_queue
            # only control the first unit!
            break
        return lux_action


class SimpleDiscreteController(Controller):
    def __init__(self, env_cfg: EnvConfig) -> None:
        """
        A simple controller that uses a discrete action parameterization for Lux AI S2. It includes

        For units
        - 4 cardinal direction movement (4 dims)
        - a move center no-op action (1 dim)
        - transfer action for each combination of the (4 cardinal directions plus center) x (resource type or power) (5*5 = 25 dims)
        - pickup action for each resource (5 dims)
        - dig action (1 dim)

        For factories
        - all actions (build light, heavy, or water) (3 dims)


        It does not include
        - self destruct action
        - recharge action
        - planning (via actions executing multiple times or repeating actions)

        Sampling from this controller will always result in a valid action, albeit sometimes disastrous
        """
        self.env_cfg = env_cfg
        self.move_act_dims = 5
        self.transfer_act_dims = 5 * 5
        self.pickup_act_dims = 5
        self.dig_act_dims = 1
        # self.self_destruct_act_dims = 1
        # self.recharge_act_dims = 1
        self.factory_act_dims = 3  # 0 = light, 1 = heavy, 2 = water

        self.move_dim_high = self.move_act_dims
        self.transfer_dim_high = self.move_dim_high + self.transfer_act_dims
        self.pickup_dim_high = self.transfer_dim_high + self.pickup_act_dims
        self.dig_dim_high = self.pickup_dim_high + self.dig_act_dims

        self.factory_dim_high = 3  # self.dig_dim_high + self.factory_act_dims

        total_act_dims = self.factory_dim_high
        # action_space = spaces.Discrete(total_act_dims)
        action_space = spaces.Box(
            0, 1, shape=(env_cfg.map_size, env_cfg.map_size, total_act_dims)
        )

        super().__init__(action_space)

    # note that all the _is_x_action are meant to be called in a if, elseif... cascade/waterfall
    # to understand how _get_x_action works to map the parameterization back to the original action space see luxai_s2/actions.py
    def _is_move_action(self, id):
        return id < self.move_dim_high

    def _get_move_action(self, id):
        return np.array([0, id, 0, 0, 0, 1])

    def _is_transfer_action(self, id):
        return id < self.transfer_dim_high

    def _get_transfer_action(self, id):
        id = id - self.move_dim_high
        transfer_dir = id % 5
        resource_type = id // 5
        return np.array(
            [1, transfer_dir, resource_type, self.env_cfg.max_transfer_amount, 0, 1]
        )

    def _is_pickup_action(self, id):
        return id < self.pickup_dim_high

    def _get_pickup_action(self, id):
        id = id - self.transfer_dim_high
        return np.array([2, 0, id % 5, self.env_cfg.max_transfer_amount, 0, 1])

    def _is_dig_action(self, id):
        return id < self.dig_dim_high

    def _get_dig_action(self, id):
        return np.array([3, 0, 0, 0, 0, 1])

    # def _is_self_destruct_action(self, id):
    # return id < self.move_act_dims + self.transfer_act_dims + self.self_destruct_dims
    # def _get_self_destruct_action(self, id):
    #     return [2, 0, 0, 0, 0, 1]

    def action_to_lux_action(
        self, agent: str, obs: Dict[str, ObservationStateDict], action: npt.NDArray
    ):
        """
        Generate an action compatible with LuxAI_S2 engine for a single player
        """
        shared_obs = obs["player_0"]
        lux_action = dict()
        factories = shared_obs["factories"][agent]
        units = shared_obs["units"][agent]
        for unit_id in units.keys():
            unit = units[unit_id]
            pos = unit["pos"]
            action_here = action[pos[0], pos[1]]
            unit_related_action = action_here[
                : -self.factory_act_dims
            ]  # assuming factory action is always the final few dimensions
            choice = unit_related_action.argmax()
            action_queue = []
            # if self._is_move_action(choice):
            #     action_queue = [self._get_move_action(choice)]
            # elif self._is_transfer_action(choice):
            #     action_queue = [self._get_transfer_action(choice)]
            # elif self._is_pickup_action(choice):
            #     action_queue = [self._get_pickup_action(choice)]
            # elif self._is_dig_action(choice):
            #     action_queue = [self._get_dig_action(choice)]

            lux_action[unit_id] = action_queue

        for unit_id in factories.keys():
            factory = factories[unit_id]
            pos = factory["pos"]

            action_here = action[pos[0], pos[1]]
            factory_related_action = action_here[
                -self.factory_act_dims :
            ]  # assuming factory action is always the final few dimensions
            choice = factory_related_action.argmax()
            lux_action[unit_id] = choice
        return lux_action



import os

TERM_COLORS = True
try:
    from termcolor import colored

    TERM_COLORS = (
        os.environ["LUX_COLORS"] == "False" if "LUX_COLORS" in os.environ else True
    )
except:
    TERM_COLORS = False
    print("termcolor not installed, skipping dependency")
    pass



import copy
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Dict, List
try:
    from typing import TypedDict    
except:
    from typing_extensions import TypedDict


class RobotStatsStateDict(TypedDict):
    LIGHT: int
    HEAVY: int


def create_robot_stats() -> RobotStatsStateDict:
    return dict(LIGHT=0, HEAVY=0)


class AllStatsStateDict(TypedDict):
    LIGHT: int
    HEAVY: int
    FACTORY: int


def create_all_stats() -> AllStatsStateDict:
    return dict(LIGHT=0, HEAVY=0, FACTORY=0)


class GenerationStatsStateDict(TypedDict):
    power: AllStatsStateDict
    water: int
    metal: int
    ore: RobotStatsStateDict  # amount dug out by HEAVY or LIGHT
    ice: RobotStatsStateDict  # amount dug out by HEAVY or LIGHT
    lichen: int  # amount grown
    built: RobotStatsStateDict  # amount built


def create_generation_stats() -> GenerationStatsStateDict:
    return dict(
        power=create_all_stats(),
        water=0,
        metal=0,
        ore=create_robot_stats(),
        ice=create_robot_stats(),
        lichen=0,
        built=create_robot_stats(),
    )


class ConsumptionStatsStateDict(TypedDict):
    power: AllStatsStateDict
    water: int
    metal: int


def create_consumption_stats() -> ConsumptionStatsStateDict:
    return dict(
        power=create_all_stats(),
        water=0,
        metal=0,
        ore=create_robot_stats(),
        ice=create_robot_stats(),
    )


class TransferStatsStateDict(TypedDict):
    power: int
    water: int
    metal: int
    ice: int
    ore: int


def create_transfer_pickup_stats():
    return dict(power=0, water=0, metal=0, ice=0, ore=0)


class PickUpStatsStateDict(TypedDict):
    power: int
    water: int
    metal: int
    ice: int
    ore: int


class DestroyedStatsStateDict(TypedDict):
    FACTORY: int
    HEAVY: int
    LIGHT: int
    rubble: RobotStatsStateDict
    lichen: RobotStatsStateDict


def create_destroyed_stats():
    return dict(
        FACTORY=0,
        HEAVY=0,
        LIGHT=0,
        rubble=create_robot_stats(),
        lichen=create_robot_stats(),
    )


class StatsStateDict(TypedDict):
    consumption: ConsumptionStatsStateDict
    generation: GenerationStatsStateDict
    action_queue_updates_success: int
    action_queue_updates_total: int
    destroyed: DestroyedStatsStateDict
    transfer: TransferStatsStateDict
    pickup: PickUpStatsStateDict


## TODO add collision stats
## TODO add above as a wrapper for jax env?


def create_empty_stats() -> StatsStateDict:
    stats: StatsStateDict = dict()
    stats["action_queue_updates_total"] = 0
    stats["action_queue_updates_success"] = 0
    stats["consumption"] = create_consumption_stats()
    stats["destroyed"] = create_destroyed_stats()
    stats["generation"] = create_generation_stats()
    stats["pickup"] = create_transfer_pickup_stats()
    stats["transfer"] = create_transfer_pickup_stats()
    return stats



import copy
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Dict, List
try:
    from typing import TypedDict    
except:
    from typing_extensions import TypedDict

import numpy as np
import numpy.typing as npt

from luxai_s2.actions import format_action_vec
from luxai_s2.config import EnvConfig
from luxai_s2.factory import Factory, FactoryStateDict
from luxai_s2.map.board import Board, BoardStateDict
from luxai_s2.map_generator.generator import GameMap
from luxai_s2.state.stats import StatsStateDict
from luxai_s2.team import Team, TeamStateDict
from luxai_s2.unit import FactionTypes, Unit, UnitCargo, UnitStateDict, UnitType


class SparseBoardStateDict(TypedDict):
    rubble: Dict[str, int]
    lichen: Dict[str, int]
    lichen_strains: Dict[str, int]
    factories_per_team: int


class DeltaObservationStateDict(TypedDict):
    units: int
    teams: Dict[str, TeamStateDict]
    factories: Dict[str, Dict[str, FactoryStateDict]]
    board: SparseBoardStateDict
    real_env_steps: int
    global_id: int


class ObservationStateDict(TypedDict):
    units: Dict[str, Dict[str, UnitStateDict]]
    teams: Dict[str, TeamStateDict]
    factories: Dict[str, Dict[str, FactoryStateDict]]
    board: BoardStateDict
    real_env_steps: int
    global_id: int


@dataclass
class State:
    seed_rng: np.random.RandomState
    seed: int
    env_steps: int
    env_cfg: EnvConfig
    board: Board = None
    units: Dict[str, Dict[str, Unit]] = field(default_factory=dict)
    factories: Dict[str, Dict[str, Factory]] = field(default_factory=dict)
    teams: Dict[str, Team] = field(default_factory=dict)
    global_id: int = 0
    stats: Dict[str, StatsStateDict] = field(default_factory=dict)

    @property
    def real_env_steps(self):
        """
        the actual env step in the environment, which subtracts the time spent bidding and placing factories
        """
        if self.env_cfg.BIDDING_SYSTEM:
            # + 1 for bidding step, * 2 for total factories placed by two teams
            return self.env_steps - (self.board.factories_per_team * 2 + 1)
        else:
            return self.env_steps

    def real_env_steps_to_env_steps(self, real_env_steps):
        if self.env_cfg.BIDDING_SYSTEM:
            # + 1 for bidding step, * 2 for total factories placed by two teams
            return real_env_steps + (self.board.factories_per_team * 2 + 1)
        else:
            return real_env_steps

    # can be somewhat slow, can we parallelize this easily?
    @staticmethod
    def generate_unit_data(units_dict: Dict[str, Dict[str, Unit]]):
        units = dict()
        for team in units_dict:
            units[team] = dict()
            for unit in units_dict[team].values():
                state_dict = unit.state_dict()
                units[team][unit.unit_id] = state_dict
        return units

    @staticmethod
    def generate_team_data(teams_dict):
        teams = dict()
        for k, v in teams_dict.items():
            teams[k] = v.state_dict()
        return teams

    @staticmethod
    def generate_factory_data(factories_dict):
        factories = dict()
        for team in factories_dict:
            factories[team] = dict()
            for factory in factories_dict[team].values():
                state_dict = factory.state_dict()
                factories[team][factory.unit_id] = state_dict
        return factories

    def get_obs(self) -> ObservationStateDict:
        units = State.generate_unit_data(self.units)
        teams = State.generate_team_data(self.teams)
        factories = State.generate_factory_data(self.factories)
        board = self.board.state_dict()
        return dict(
            units=units,
            teams=teams,
            factories=factories,
            board=board,
            real_env_steps=self.real_env_steps,
            global_id=self.global_id,
        )

    def get_compressed_obs(self):
        # return everything on turn 0
        if self.env_steps == 0:
            return self.get_obs()
        data = self.get_obs()
        # convert lichen and lichen strains to sparse matrix format?
        del data["board"]["ore"]
        del data["board"]["ice"]
        if self.real_env_steps >= 0:
            del data["board"]["valid_spawns_mask"]
        return data

    def get_change_obs(
        self, prev_state: ObservationStateDict
    ) -> DeltaObservationStateDict:
        """
        returns sparse dicts for large matrices of where values change only by comparing against a given previous observation/state
        """
        data = self.get_compressed_obs()

        data["board"]["rubble"] = dict()
        data["board"]["lichen"] = dict()
        data["board"]["lichen_strains"] = dict()
        change_indices = np.argwhere(self.board.rubble != prev_state["board"]["rubble"])
        for ind in change_indices:
            x, y = ind[0], ind[1]
            data["board"]["rubble"][f"{x},{y}"] = self.board.rubble[x, y]
        change_indices = np.argwhere(self.board.lichen != prev_state["board"]["lichen"])
        for ind in change_indices:
            x, y = ind[0], ind[1]
            data["board"]["lichen"][f"{x},{y}"] = self.board.lichen[x, y]
        change_indices = np.argwhere(
            self.board.lichen_strains != prev_state["board"]["lichen_strains"]
        )
        for ind in change_indices:
            x, y = ind[0], ind[1]
            data["board"]["lichen_strains"][f"{x},{y}"] = self.board.lichen_strains[
                x, y
            ]
        return data

    @staticmethod
    def accumulate_board_changes(board: Board, board_change_observations: List):
        # Accumulates the delta changes to a board from change observations.
        # Should be used with `from_obs` if the obs given is missing initial observation info
        for obs in board_change_observations:
            for item in ["rubble", "lichen", "lichen_strains"]:
                for k, v in obs[item].items():
                    k = k.split(",")
                    x, y = int(k[0]), int(k[1])
                    board.__getattribute__(item)[x, y] = v

    @classmethod
    def from_obs(cls, obs: ObservationStateDict, env_cfg: EnvConfig):
        real_env_steps = obs["real_env_steps"]
        factories_per_team = obs["board"]["factories_per_team"]
        if env_cfg.BIDDING_SYSTEM:
            # + 1 for bidding step, * 2 for total factories placed by two teams
            env_steps = real_env_steps + (factories_per_team * 2 + 1)
        else:
            env_steps = real_env_steps

        teams = dict()
        for agent in obs["teams"]:
            team_data = obs["teams"][agent]
            faction = FactionTypes[team_data["faction"]]
            team = Team(team_id=team_data["team_id"], agent=agent, faction=faction)
            team.bid = team_data["bid"]
            team.init_water = team_data["water"]
            team.init_metal = team_data["metal"]
            team.factories_to_place = team_data["factories_to_place"]
            team.factory_strains = team_data["factory_strains"]
            team.place_first = team_data["place_first"]
            teams[agent] = team

        existing_map = GameMap(
            np.array(obs["board"]["rubble"]),
            np.array(obs["board"]["ice"]),
            np.array(obs["board"]["ore"]),
            None,
        )

        board = Board(seed=0, env_cfg=env_cfg, existing_map=existing_map)
        board.factories_per_team = factories_per_team

        units = dict()
        for agent in obs["units"]:
            units[agent] = dict()
            for unit_id in obs["units"][agent]:
                unit_data = obs["units"][agent][unit_id]
                cargo = UnitCargo(**unit_data["cargo"])
                unit = Unit(
                    teams[agent], UnitType[unit_data["unit_type"]], unit_id, env_cfg
                )
                unit.pos.pos = np.array(unit_data["pos"])
                unit.cargo = cargo
                unit.power = unit_data["power"]
                unit.action_queue = [
                    format_action_vec(a) for a in unit_data["action_queue"]
                ]

                units[agent][unit_id] = unit
                board.units_map[board.pos_hash(unit.pos)].append(unit)

        factory_occupancy_map = np.ones_like(obs["board"]["rubble"], dtype=int) * -1
        factories = dict()
        for agent in obs["factories"]:
            factories[agent] = dict()
            for unit_id in obs["factories"][agent]:
                f_data = obs["factories"][agent][unit_id]
                cargo = UnitCargo(**f_data["cargo"])
                factory = Factory(teams[agent], unit_id, f_data["strain_id"])
                factory.cargo = cargo
                factory.power = f_data["power"]
                factory.pos.pos = np.array(f_data["pos"])
                factories[agent][unit_id] = factory

                # code below mimics env.py's add_factory function
                factory_occupancy_map[factory.pos_slice] = factory.num_id

                board.factory_map[board.pos_hash(factory.pos)] = factory

                invalid_spawn_indices = factory.min_dist_slice
                for x, y in invalid_spawn_indices:
                    if (
                        x < 0
                        or y < 0
                        or x >= board.rubble.shape[0]
                        or y >= board.rubble.shape[1]
                    ):
                        continue
                    board.valid_spawns_mask[x, y] = False

                # bottom 3 might not be necessary
                board.rubble[factory.pos_slice] = 0
                board.ice[factory.pos_slice] = 0
                board.ore[factory.pos_slice] = 0

        # update maps inside board
        board.factory_occupancy_map = factory_occupancy_map

        # generate state from full observation
        return cls(
            seed_rng=np.random.RandomState(seed=0),  # unused
            seed=0,  # unused
            env_steps=env_steps,
            env_cfg=env_cfg,
            board=board,
            units=units,
            teams=teams,
            factories=factories,
            global_id=obs["global_id"],
            stats=dict(),  # not saved
        )



from turtle import color

from luxai_s2.globals import TERM_COLORS

try:
    from termcolor import colored
except:
    pass
import os


class Logger:
    """
    A basic logger

    verbosity - how verbose the logger is
        0 : Log nothing
        1 : Log errors
        2 : Log warnings
        3 : Log everything
    """

    def __init__(self, identifier: str, verbosity: int = 1) -> None:
        self.verbosity = verbosity
        self.identifier = identifier

    def _print(self, msg: str, color: str, end: str = "\n"):
        if self.identifier != "":
            msg = f"{self.identifier}: {msg}"
        if TERM_COLORS:
            print(colored(msg, color), end=end)
        else:
            print(msg, end=end)

    def err(self, msg: str, end: str = "\n"):
        if self.verbosity >= 1:
            self._print(msg, "red", end=end)

    def warn(self, msg: str, end: str = "\n"):
        if self.verbosity >= 2:
            self._print(msg, "yellow", end=end)

    def info(self, msg: str, end: str = "\n"):
        if self.verbosity >= 3:
            self._print(msg, "cyan", end=end)



import math
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Callable

import numpy as np

from luxai_s2.config import EnvConfig

if TYPE_CHECKING:
    from luxai_s2.factory import Factory
    from luxai_s2.state import State

import luxai_s2.unit as luxai_unit
from luxai_s2.map.position import Position


# (0 = move, 1 = transfer X amount of R, 2 = pickup X amount of R, 3 = dig, 4 = self destruct, 5 = recharge X)
class Action:
    def __init__(self, act_type: str) -> None:
        self.act_type = act_type
        self.n = 1  # number of times to execute the action
        self.power_cost = 0
        self.repeat = False  # whether to put action to the back of the action queue

    def state_dict(self):
        raise NotImplementedError("")


class FactoryBuildAction(Action):
    def __init__(self, unit_type: luxai_unit.UnitType) -> None:
        super().__init__("factory_build")
        self.unit_type = unit_type
        self.power_cost = 0

    def state_dict(self):
        if self.unit_type == luxai_unit.UnitType.LIGHT:
            return 0
        return 1

    def __str__(self) -> str:
        return f"{self.act_type} {self.unit_type.name}"


class FactoryWaterAction(Action):
    def __init__(self) -> None:
        super().__init__("factory_water")
        self.water_cost = None
        self.power_cost = 0

    def state_dict(self):
        return 2

    def __str__(self) -> str:
        return f"{self.act_type}: (cost: {self.water_cost}"


direction_to_name = ["center", "up", "right", "down", "left"]
resource_to_name = ["ice", "ore", "water", "metal", "power"]


class MoveAction(Action):
    def __init__(
        self, move_dir: int, dist: int = 1, repeat: bool = False, n: int = 1
    ) -> None:
        super().__init__("move")
        # a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)
        self.move_dir = move_dir
        self.dist = dist
        self.repeat = repeat
        self.n = n
        self.power_cost = 0

    def state_dict(self):
        return np.array([0, self.move_dir, 0, 0, self.repeat, self.n])

    def __str__(self) -> str:
        return f"{self.act_type} {direction_to_name[self.move_dir]} (n: {self.n}, r: {self.repeat})"


class TransferAction(Action):
    def __init__(
        self,
        transfer_dir: int,
        resource: int,
        transfer_amount: int,
        repeat: bool = False,
        n: int = 1,
    ) -> None:
        super().__init__("transfer")
        # a[2] = R = resource type (0 = ice, 1 = ore, 2 = water, 3 = metal, 4 power)
        self.transfer_dir = transfer_dir
        self.resource = resource
        self.transfer_amount = transfer_amount
        self.repeat = repeat
        self.n = n
        self.power_cost = 0

    def state_dict(self):
        return np.array(
            [
                1,
                self.transfer_dir,
                self.resource,
                self.transfer_amount,
                self.repeat,
                self.n,
            ]
        )

    def __str__(self) -> str:
        return f"{self.act_type} {resource_to_name[self.resource]} {direction_to_name[self.transfer_dir]} (n: {self.n}, r: {self.repeat})"


class PickupAction(Action):
    def __init__(
        self, resource: int, pickup_amount: int, repeat: bool = False, n: int = 1
    ) -> None:
        super().__init__("pickup")
        # a[2] = R = resource type (0 = ice, 1 = ore, 2 = water, 3 = metal, 4 power)
        self.resource = resource
        self.pickup_amount = pickup_amount
        self.repeat = repeat
        self.n = n
        self.power_cost = 0

    def state_dict(self):
        return np.array([2, 0, self.resource, self.pickup_amount, self.repeat, self.n])

    def __str__(self) -> str:
        return f"{self.act_type} {self.pickup_amount} {resource_to_name[self.resource]} (n: {self.n}, r: {self.repeat})"


class DigAction(Action):
    def __init__(self, repeat: bool = False, n: int = 1) -> None:
        super().__init__("dig")
        self.repeat = repeat
        self.n = n
        self.power_cost = 0

    def state_dict(self):
        return np.array([3, 0, 0, 0, self.repeat, self.n])

    def __str__(self) -> str:
        return f"{self.act_type} (n: {self.n}, r: {self.repeat})"


class SelfDestructAction(Action):
    def __init__(self, repeat: bool = False, n: int = 1) -> None:
        super().__init__("self_destruct")
        self.repeat = repeat
        self.n = n
        self.power_cost = 0

    def state_dict(self):
        return np.array([4, 0, 0, 0, self.repeat, self.n])

    def __str__(self) -> str:
        return f"{self.act_type} (n: {self.n}, r: {self.repeat})"


class RechargeAction(Action):
    def __init__(self, power: int, repeat: bool = False, n: int = 1) -> None:
        super().__init__("recharge")
        self.power = power
        self.repeat = repeat
        self.n = n
        self.power_cost = 0

    def state_dict(self):
        return np.array([5, 0, 0, self.power, self.repeat, self.n])

    def __str__(self) -> str:
        return f"{self.act_type} {self.power} (n: {self.n}, r: {self.repeat})"


def format_factory_action(a: int):
    if a == 0 or a == 1:
        unit_type = luxai_unit.UnitType.HEAVY
        if a == 0:
            unit_type = luxai_unit.UnitType.LIGHT
        return FactoryBuildAction(unit_type=unit_type)
    elif a == 2:
        return FactoryWaterAction()
    else:
        raise ValueError(f"Action {a} for factory is invalid")


def format_action_vec(a: np.ndarray):
    # (0 = move, 1 = transfer X amount of R, 2 = pickup X amount of R, 3 = dig, 4 = self destruct, 5 = recharge X)
    a_type = a[0]
    if a_type == 0:
        return MoveAction(a[1], dist=1, repeat=a[4], n=a[5])
    elif a_type == 1:
        return TransferAction(a[1], a[2], a[3], repeat=a[4], n=a[5])
    elif a_type == 2:
        return PickupAction(a[2], a[3], repeat=a[4], n=a[5])
    elif a_type == 3:
        return DigAction(repeat=a[4], n=a[5])
    elif a_type == 4:
        return SelfDestructAction(repeat=a[4], n=a[5])
    elif a_type == 5:
        return RechargeAction(a[3], repeat=a[4], n=a[5])
    else:
        raise ValueError(f"Action {a} is invalid type, {a[0]} is not valid")


# a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)
move_deltas = np.array([[0, 0], [0, -1], [1, 0], [0, 1], [-1, 0]])


def validate_actions(env_cfg: EnvConfig, state: "State", actions_by_type, verbose=1):
    """
    validates actions and logs warnings for any invalid actions. Invalid actions are subsequently not evaluated and
    won't count towards the `n` value of actions that execute multiple times.
    """
    actions_by_type_validated = defaultdict(list)
    valid_action = True

    def invalidate_action(msg):
        nonlocal valid_action
        valid_action = False
        if verbose > 0:
            print(f"{state.real_env_steps}: {msg}")

    for unit, transfer_action in actions_by_type["transfer"]:
        valid_action = True
        unit: luxai_unit.Unit
        transfer_action: TransferAction
        transfer_pos: Position = unit.pos + move_deltas[transfer_action.transfer_dir]
        if transfer_action.resource > 4 or transfer_action.resource < 0:
            invalidate_action(
                f"Invalid Transfer Action for unit {unit}, transferring invalid resource id {transfer_action.resource}"
            )
            continue
        if (
            transfer_pos.x < 0
            or transfer_pos.y < 0
            or transfer_pos.x >= state.board.width
            or transfer_pos.y >= state.board.height
        ):
            invalidate_action(
                f"Invalid Transfer action for unit {unit} - Tried to transfer to {transfer_pos} which is off the map"
            )
            continue
        # if transfer_action.transfer_amount < 0: do not need to check as action space permits range of [0, max_transfer_amount] anyway
        resource_id = transfer_action.resource
        amount = transfer_action.transfer_amount

        if valid_action:
            actions_by_type_validated["transfer"].append((unit, transfer_action))

    for unit, dig_action in actions_by_type["dig"]:
        valid_action = True
        dig_action: DigAction
        dig_cost = unit.unit_cfg.DIG_COST
        if dig_cost > unit.power:
            invalidate_action(
                f"Invalid Dig Action for unit {unit} - Tried to dig requiring {dig_cost} power"
            )
            continue
        # verify not digging over a factory which is not allowed
        if state.board.factory_occupancy_map[unit.pos.x, unit.pos.y] != -1:
            invalidate_action(
                f"Invalid Dig Action for unit {unit} - Tried to dig on top of a factory"
            )
            continue
        if valid_action:
            actions_by_type_validated["dig"].append((unit, dig_action))

    for unit, pickup_action in actions_by_type["pickup"]:
        valid_action = True
        pickup_action: PickupAction
        unit: luxai_unit.Unit
        factory = state.board.get_factory_at(state, unit.pos)
        if factory is None:
            invalidate_action(f"No factory to pickup from for unit {unit}")
            continue
        if valid_action:
            actions_by_type_validated["pickup"].append((unit, pickup_action))

    for unit, move_action in actions_by_type["move"]:
        valid_action = True
        move_action: MoveAction
        target_pos: Position = (
            unit.pos + move_action.dist * move_deltas[move_action.move_dir]
        )
        if (
            target_pos.x < 0
            or target_pos.y < 0
            or target_pos.x >= state.board.width
            or target_pos.y >= state.board.height
        ):
            invalidate_action(
                f"Invalid movement action for unit {unit} - Tried to move to {target_pos} which is off the map"
            )
            continue
        if state.board.factory_occupancy_map[target_pos.x, target_pos.y] != -1:
            factory_id = state.board.factory_occupancy_map[target_pos.x, target_pos.y]
            if f"factory_{factory_id}" not in state.factories[unit.team.agent]:
                # if there is a factory but not same team
                invalidate_action(
                    f"Invalid movement action for unit {unit} - Tried to move to {target_pos} which is on an opponent factory"
                )
                continue
        rubble = state.board.rubble[target_pos.x, target_pos.y]
        power_required = (
            0 if move_action.move_dir == 0 else unit.move_power_cost(rubble)
        )

        if power_required > unit.power:
            invalidate_action(
                f"Invalid movement action for unit {unit} - Tried to move to {target_pos} requiring {power_required} power"
            )
            continue
        if valid_action:
            move_action.power_cost = power_required
            actions_by_type_validated["move"].append((unit, move_action))

    for unit, self_destruct_action in actions_by_type["self_destruct"]:
        valid_action = True
        self_destruct_action: SelfDestructAction
        power_required = unit.unit_cfg.SELF_DESTRUCT_COST
        if power_required > unit.power:
            invalidate_action(
                f"Invalid self destruct action for unit {unit} - Tried to self destruct requiring {power_required} power"
            )
            continue
        if valid_action:
            self_destruct_action.power_cost = power_required
            actions_by_type_validated["self_destruct"].append(
                (unit, self_destruct_action)
            )

    for unit, recharge_action in actions_by_type["recharge"]:
        actions_by_type_validated["recharge"].append((unit, recharge_action))

    for factory, build_action in actions_by_type["factory_build"]:
        valid_action = True
        build_action: FactoryBuildAction
        factory: "Factory"

        unit_cfg = env_cfg.ROBOTS[build_action.unit_type.name]
        if factory.cargo.metal < unit_cfg.METAL_COST:
            invalidate_action(
                f"Invalid factory build action for factory {factory} - Insufficient metal, factory has {factory.cargo.metal}, but requires {unit_cfg.METAL_COST} to build {build_action.unit_type}"
            )
            continue
        power_required = unit_cfg.POWER_COST
        if factory.power < power_required:
            invalidate_action(
                f"Invalid factory build action for factory {factory} - Insufficient power, factory has {factory.power}, but requires {power_required} power"
            )
            continue
        if valid_action:
            build_action.power_cost = power_required
            actions_by_type_validated["factory_build"].append((factory, build_action))
        pass

    for factory, water_action in actions_by_type["factory_water"]:
        valid_action = True
        water_cost = factory.water_cost(env_cfg)
        if water_cost > factory.cargo.water:
            invalidate_action(
                f"Invalid factory water action for factory {factory} - Insufficient water, factory has {factory.cargo.water}, but requires {water_cost} to water lichen"
            )
            continue
        if valid_action:
            actions_by_type_validated["factory_water"].append((factory, water_action))

    return actions_by_type_validated



import numpy as np


def to_json(obj):
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, list) or isinstance(obj, tuple):
        return [to_json(s) for s in obj]
    elif isinstance(obj, dict):
        out = {}
        for k in obj:
            out[k] = to_json(obj[k])
        return out
    else:
        return obj



import os
import os.path as osp
from typing import List

from luxai_s2.replay import decode_replay_file
from luxai_s2.state import State


def replay_trajectory(replay_file: str):
    decoded = decode_replay_file(replay_file)
    if "observations" in decoded:
        print("replay file already contains observations, there is nothing to do")
        return replay_file



import random
from typing import Any, Dict, List

import gym
import numpy as np
from gym import spaces

from luxai_s2.config import EnvConfig
from luxai_s2.factory import Factory
from luxai_s2.team import FactionTypes
from luxai_s2.unit import Unit


class FactionString(gym.Space):
    def __init__(
        self,
    ):
        self.valid_factions = ["Null"] + [x.name for x in FactionTypes]

    def sample(self):
        return random.choice(self.valid_factions)

    def contains(self, x):
        return type(x) == "str" and x in self.valid_factions


class PartialDict(spaces.Dict):
    def contains(self, x: Any) -> bool:
        if not isinstance(x, dict) or len(x) > len(self.spaces):
            return False, "invalid"
        for k, v in x.items():
            if k in self.spaces:
                space = self.spaces[k]
                if not space.contains(v):
                    return False, f"invalid action {v} for {k}"
            else:
                return False, f"{k} is not on your team or does not exist"
        return True, None


class ActionsQueue(spaces.Space):
    def __init__(self, action_space: spaces.Space, max_length: int) -> None:
        super().__init__((), float)
        self.max_length = max_length
        self.action_space = action_space

    def sample(self):
        queue_size = self.np_random.randint(0, self.max_length) + 1
        action_q = []
        for _ in range(queue_size):
            action_q.append(self.action_space.sample())
        return action_q

    def contains(self, x: Any) -> bool:
        if isinstance(x, list) and len(x) > self.max_length:
            return False
        elif (
            isinstance(x, np.ndarray) and len(x.shape) == 2 and len(x) > self.max_length
        ):
            return False
        elif isinstance(x, np.ndarray) and len(x.shape) == 1:
            x = [x]
        # if (not isinstance(x, list) and not isinstance(x, np.ndarray)) or len(x) > self.max_length:
        # return False
        for a in x:
            # fix issue where multidiscrete space does not do 100% check on type of value
            try:
                if not self.action_space.contains(a):
                    return False
            except:
                return False
        return True


def get_act_space_init(config: EnvConfig, agent: str):
    # Get action space for turn 0 initialization
    act_space = dict()
    act_space["faction"] = FactionString()
    act_space["spawns"] = spaces.Box(
        low=0, high=config.map_size, shape=(config.MAX_FACTORIES, 2), dtype=int
    )
    return spaces.Dict(act_space)


def get_act_space_bid(config: EnvConfig, agent: str):
    act_space = dict()
    act_space["faction"] = FactionString()
    act_space["bid"] = spaces.Discrete(100000)
    return spaces.Dict(act_space)


def get_act_space_placement(config: EnvConfig, agent: str):
    # Get action space for turn 0 initialization
    act_space = dict()
    act_space["spawn"] = spaces.Box(low=0, high=config.map_size, shape=(2,), dtype=int)
    act_space["water"] = spaces.Discrete(100000)
    act_space["metal"] = spaces.Discrete(100000)
    return spaces.Dict(act_space)


def get_act_space(
    units: Dict[str, Dict[str, Unit]],
    factories: Dict[str, Dict[str, Factory]],
    config: EnvConfig,
    agent: str,
):
    act_space = dict()

    # for consistency, every action space per unit is fixed, makes it easier to work out of the box.
    # we further annotate dimensions, in many places for clarity

    # for those who are programming rule-based bots, we provide helper functions to generate the action vector to store in an actions dict, as well as generate
    # a human readable version of an action vector

    # TODO - verify speed of building action spaces like this.
    for u in units[agent].values():
        # Each action for any mobile unit, (light, heavy) is an array A of shape (max_queue, 5)
        # if max_queue is 1, we remove this dimension
        # let a be some element in A
        # Then

        # a[0] = action type
        # (0 = move, 1 = transfer X amount of R, 2 = pickup X amount of R, 3 = dig, 4 = self destruct, 5 = recharge X)

        # a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)

        # a[2] = R = resource type (0 = ice, 1 = ore, 2 = water, 3 = metal, 4 power)

        # a[3] = X, amount of resources transferred or picked up if action is transfer or pickup.
        # If action is recharge, it is how much energy to store before executing the next action in queue

        # a[4] = 0 or 1 (false or true). If True, then action is placed to the back of the action queue after it has been exhausted

        # a[5] = X, number of times to execute this action before exhausting it and removing it from the front of the action queue. Minimum is 1.
        act_space[u.unit_id] = ActionsQueue(
            spaces.Box(
                low=np.array([0, 0, 0, 0, 0, 1]),
                high=np.array([5, 4, 4, config.max_transfer_amount + 1, 1, 9999]),
                shape=(6,),
                dtype=np.int64,
            ),
            config.UNIT_ACTION_QUEUE_SIZE,
        )

    for factory in factories[agent].values():
        # action type (0 = build light robot, 1 = build heavy robot, 2 = water lichen)
        act_space[factory.unit_id] = spaces.Discrete(3)

    return PartialDict(act_space)



from math import prod
from pathlib import Path

import numpy as np
import pygame

from luxai_s2.env import LuxAI_S2

if __name__ == "__main__":
    env = LuxAI_S2()
    N = 100
    Path("./map_samples").mkdir(parents=True, exist_ok=True)
    stats = dict(rubble_count=[], ice=[], ore=[])
    for seed in range(10000, 10000 + N):
        env.reset(seed)
        env.render()
        env.py_visualizer.update_scene(env.state)

        pygame.image.save(env.py_visualizer.screen, f"map_samples/{seed}.jpg")
        stats["ice"] += [env.state.board.ice.sum()]
        stats["ore"] += [env.state.board.ore.sum()]
        stats["rubble_count"] += [env.state.board.rubble.sum()]

    for k in stats:
        stats[k] = np.mean(stats[k])
    stats["rubble_count"] /= prod(env.state.board.rubble.shape)
    print(stats)



import asyncio
import dataclasses
import json
import os.path as osp
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional

import gym
import numpy as np
from luxai_runner.bot import Bot
from luxai_runner.logger import Logger
from luxai_runner.utils import to_json


@dataclass
class ReplayConfig:
    save_format: str = "json"
    compressed_obs: bool = False


@dataclass
class EpisodeConfig:
    players: List[str]
    env_cls: Callable[[Any], gym.Env]
    seed: Optional[int] = None
    env_cfg: Optional[Any] = dict
    verbosity: Optional[int] = 1
    render: Optional[bool] = True
    save_replay_path: Optional[str] = None
    replay_options: ReplayConfig = ReplayConfig()


class Episode:
    def __init__(self, cfg: EpisodeConfig) -> None:
        self.cfg = cfg
        self.env = cfg.env_cls(**cfg.env_cfg)
        self.log = Logger(identifier="Episode", verbosity=cfg.verbosity)
        self.seed = cfg.seed if cfg.seed is not None else np.random.randint(9999999)
        self.players = cfg.players

    def save_replay(self, replay):
        save_format = self.cfg.replay_options.save_format
        if save_format not in ["json", "html"]:
            raise ValueError(f"{save_format} is not a valid save format")

        replay["observations"] = [to_json(x) for x in replay["observations"]]
        replay["actions"] = [to_json(x) for x in replay["actions"]]
        del replay["dones"]
        del replay["rewards"]

        ext = f".{save_format}"

        from pathlib import Path

        dir_name = osp.dirname(self.cfg.save_replay_path)
        if dir_name != "":
            Path(dir_name).mkdir(parents=True, exist_ok=True)

        if self.cfg.save_replay_path[-5:] == ext:
            ext = ""

        with open(f"{self.cfg.save_replay_path}{ext}", "w") as f:
            if save_format == "json":
                json.dump(replay, f)
            else:
                f.write(
                    f"""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <link rel="icon" type="image/svg+xml" href="https://s2vis.lux-ai.org/eye.svg" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />

    <title>Lux Eye S2</title>

    <script>
window.episode = {json.dumps(replay)};
    </script>

    <script type="module" crossorigin src="https://s2vis.lux-ai.org/index.js"></script>
  </head>
  <body>
    <div id="root"></div>
  </body>
</html>
                """.strip()
                )

    async def run(self):
        if len(self.players) != 2:
            raise ValueError("Must provide two paths.")
        # Start agents
        players: Dict[str, Bot] = dict()
        start_tasks = []
        save_replay = self.cfg.save_replay_path is not None
        for i in range(2):
            player = Bot(self.players[i], f"player_{i}", i, verbose=self.log.verbosity)
            player.proc.log.identifier = player.log.identifier
            players[player.agent] = player
            start_tasks += [player.proc.start()]
        await asyncio.wait(start_tasks, return_when=asyncio.ALL_COMPLETED)

        obs = self.env.reset(seed=self.seed)
        env_cfg = self.env.state.env_cfg
        state_obs = self.env.state.get_compressed_obs()
        obs = to_json(state_obs)

        if self.cfg.render:
            self.env.render()
            time.sleep(0.2)
        game_done = False
        rewards, dones, infos = dict(), dict(), dict()
        for agent in players:
            rewards[agent] = 0
            dones[agent] = 0
            infos[agent] = dict(
                # turn 0 provide configurations
                env_cfg=dataclasses.asdict(env_cfg)
            )

        if save_replay:
            replay = dict(observations=[], actions=[], dones=[], rewards=[])
            if self.cfg.replay_options.compressed_obs:
                replay["observations"].append(state_obs)
            else:
                replay["observations"].append(self.env.state.get_obs())

        i = 0
        while not game_done:
            i += 1
            # print("===", self.env.env_steps)
            actions = dict()

            agent_ids = []
            action_coros = []
            for player in players.values():
                action = player.step(
                    obs, self.env.env_steps, rewards[agent], infos[agent]
                )
                action_coros += [action]
                agent_ids += [player.agent]
            resolved_actions = await asyncio.gather(*action_coros)
            for agent_id, action in zip(agent_ids, resolved_actions):
                try:
                    for k in action:
                        if type(action[k]) == list:
                            action[k] = np.array(action[k])
                    actions[agent_id] = action
                except:
                    if self.cfg.verbosity > 0:
                        if action is None:
                            print(
                                f"{agent_id} sent a invalid action {action}. Agent likely errored out somewhere, check above for stderr logs"
                            )
                        else:
                            print(f"{agent_id} sent a invalid action {action}")
                    actions[agent_id] = None
            new_state_obs, rewards, dones, infos = self.env.step(actions)
            change_obs = self.env.state.get_change_obs(state_obs)
            state_obs = new_state_obs["player_0"]
            obs = to_json(change_obs)
            if save_replay:
                if self.cfg.replay_options.compressed_obs:
                    replay["observations"].append(change_obs)
                else:
                    replay["observations"].append(self.env.state.get_obs())
                replay["actions"].append(actions)
                replay["rewards"].append(rewards)
                replay["dones"].append(dones)

            if self.cfg.render:
                self.env.render()
                time.sleep(0.1)
            players_left = len(dones)
            for k in dones:
                if dones[k]:
                    players_left -= 1
            if players_left < 2:  # specific to lux ai s2
                game_done = True
        self.log.info(f"Final Scores: {rewards}")
        if save_replay:
            self.save_replay(replay)

        for player in players.values():
            await player.proc.cleanup()

        return rewards



# A quick visualizer to aid in map generating algorithms.
try:
    import pygame
except:
    pass
import numpy as np


def viz(game_map, screen=None):
    N = min(500 // game_map.width, 750 // game_map.height)
    if screen is None:
        screen = pygame.display.set_mode((3 * N * game_map.width, N * game_map.height))

    for x in range(game_map.width):
        for y in range(game_map.height):
            rubble_color = [255 - game_map.rubble[y][x] * 255 / 100] * 3
            ice_color = [0, 0, game_map.ice[y][x] * 255 / 100]
            ore_color = [game_map.ore[y][x] * 255 / 100, 0, 0]
            screen.fill(rubble_color, (N * x, N * y, N, N))
            screen.fill(ice_color, (N * x + N * game_map.width, N * y, N, N))
            screen.fill(ore_color, (N * x + 2 * N * game_map.width, N * y, N, N))
    pygame.display.update()



import dataclasses
from argparse import Namespace
from dataclasses import dataclass
from typing import Dict, List


def convert_dict_to_ns(x):
    if isinstance(x, dict):
        for k in x:
            x[k] = convert_dict_to_ns(x)
        return Namespace(x)


@dataclass
class UnitConfig:
    METAL_COST: int = 100
    POWER_COST: int = 500
    CARGO_SPACE: int = 1000
    BATTERY_CAPACITY: int = 1500
    CHARGE: int = 1
    INIT_POWER: int = 50
    MOVE_COST: int = 1
    RUBBLE_MOVEMENT_COST: float = 1
    DIG_COST: int = 5
    DIG_RUBBLE_REMOVED: int = 1
    DIG_RESOURCE_GAIN: int = 2
    DIG_LICHEN_REMOVED: int = 10
    SELF_DESTRUCT_COST: int = 10
    RUBBLE_AFTER_DESTRUCTION: int = 1
    ACTION_QUEUE_POWER_COST: int = 1


@dataclass
class EnvConfig:
    ## various options that can be configured if needed

    ### Variable parameters that don't affect game logic much ###
    max_episode_length: int = 1000
    map_size: int = 48
    verbose: int = 1

    # this can be disabled to improve env FPS but assume your actions are well formatted
    # During online competition this is set to True
    validate_action_space: bool = True

    ### Constants ###
    # you can only ever transfer in/out 3000 as this is the max cargo/power space.
    max_transfer_amount: int = 3000
    MIN_FACTORIES: int = 2
    MAX_FACTORIES: int = 5
    CYCLE_LENGTH: int = 50
    DAY_LENGTH: int = 30
    UNIT_ACTION_QUEUE_SIZE: int = 20  # when set to 1, then no action queue is used

    MAX_RUBBLE: int = 100
    FACTORY_RUBBLE_AFTER_DESTRUCTION: int = 50
    INIT_WATER_METAL_PER_FACTORY: int = (
        150  # amount of water and metal units given to each factory
    )
    INIT_POWER_PER_FACTORY: int = 1000

    #### LICHEN ####
    MIN_LICHEN_TO_SPREAD: int = 20
    LICHEN_LOST_WITHOUT_WATER: int = 1
    LICHEN_GAINED_WITH_WATER: int = 1
    MAX_LICHEN_PER_TILE: int = 100
    POWER_PER_CONNECTED_LICHEN_TILE: int = 1

    # cost of watering with a factory is `ceil(# of connected lichen tiles) / (this factor) + 1`
    LICHEN_WATERING_COST_FACTOR: int = 10

    #### Bidding System ####
    BIDDING_SYSTEM: bool = True

    #### Factories ####
    FACTORY_PROCESSING_RATE_WATER: int = 100
    ICE_WATER_RATIO: int = 4
    FACTORY_PROCESSING_RATE_METAL: int = 50
    ORE_METAL_RATIO: int = 5
    # game design note: Factories close to resource cluster = more resources are refined per turn
    # Then the high ice:water and ore:metal ratios encourages transfer of refined resources between
    # factories dedicated to mining particular clusters which is more possible as it is more compact

    FACTORY_CHARGE: int = 50
    FACTORY_WATER_CONSUMPTION: int = 1
    # game design note: with a positve water consumption, game becomes quite hard for new competitors.
    # so we set it to 0

    #### Collision Mechanics ####
    POWER_LOSS_FACTOR: float = 0.5

    #### Units ####
    ROBOTS: Dict[str, UnitConfig] = dataclasses.field(
        default_factory=lambda: dict(
            LIGHT=UnitConfig(
                METAL_COST=10,
                POWER_COST=50,
                INIT_POWER=50,
                CARGO_SPACE=100,
                BATTERY_CAPACITY=150,
                CHARGE=1,
                MOVE_COST=1,
                RUBBLE_MOVEMENT_COST=0.05,
                DIG_COST=5,
                SELF_DESTRUCT_COST=5,
                DIG_RUBBLE_REMOVED=2,
                DIG_RESOURCE_GAIN=2,
                DIG_LICHEN_REMOVED=10,
                RUBBLE_AFTER_DESTRUCTION=1,
                ACTION_QUEUE_POWER_COST=1,
            ),
            HEAVY=UnitConfig(
                METAL_COST=100,
                POWER_COST=500,
                INIT_POWER=500,
                CARGO_SPACE=1000,
                BATTERY_CAPACITY=3000,
                CHARGE=10,
                MOVE_COST=20,
                RUBBLE_MOVEMENT_COST=1,
                DIG_COST=60,
                SELF_DESTRUCT_COST=100,
                DIG_RUBBLE_REMOVED=20,
                DIG_RESOURCE_GAIN=20,
                DIG_LICHEN_REMOVED=100,
                RUBBLE_AFTER_DESTRUCTION=10,
                ACTION_QUEUE_POWER_COST=10,
            ),
        )
    )

    @classmethod
    def from_dict(cls, data):
        data["ROBOTS"]["LIGHT"] = UnitConfig(**data["ROBOTS"]["LIGHT"])
        data["ROBOTS"]["HEAVY"] = UnitConfig(**data["ROBOTS"]["HEAVY"])
        return cls(**data)



"""
Wrappers that allow users to insert heuristics into the environment reset and step functions
"""
from typing import Dict

import gym
import numpy as np
import numpy.typing as npt
from gym import spaces

import luxai_s2.env
from luxai_s2.env import LuxAI_S2
from luxai_s2.state import ObservationStateDict
from luxai_s2.utils import my_turn_to_place_factory
from luxai_s2.wrappers.controllers import (
    Controller,
    SimpleDiscreteController,
    SimpleSingleUnitDiscreteController,
)


class FactoryControlWrapper(gym.Wrapper):
    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)

    def step(self, action):
        return super().step(action)



import numpy as np

move_deltas = np.array([[0, 0], [0, -1], [1, 0], [0, 1], [-1, 0]])


def gen_deltas(size=3):
    start = (0, 0)
    f = [start]
    seen = set()
    r = set()
    while len(f) > 0:
        pos = f.pop(0)
        seen.add(pos)
        if abs(pos[0]) + abs(pos[1]) > size:
            break
        r.add(pos)
        for md in move_deltas:
            new_pos = (pos[0] + md[0], pos[1] + md[1])
            if new_pos not in seen:
                f.append(new_pos)
    print(r)


gen_deltas(6)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List
try:
    from typing import TypedDict    
except:
    from typing_extensions import TypedDict

import numpy as np
import numpy.typing as npt

from luxai_s2.config import EnvConfig

if TYPE_CHECKING:
    from luxai_s2.factory import Factory
    from luxai_s2.state import State

from luxai_s2.map.position import Position
from luxai_s2.map_generator.generator import GameMap
from luxai_s2.unit import Unit


class BoardStateDict(TypedDict):
    rubble: npt.NDArray[np.int_]
    ore: npt.NDArray[np.bool_]
    ice: npt.NDArray[np.bool_]
    lichen: npt.NDArray[np.int_]
    lichen_strains: npt.NDArray[np.int_]
    valid_spawns_mask: npt.NDArray[np.bool_]
    factories_per_team: int


class Board:
    def __init__(
        self, seed=None, env_cfg: EnvConfig = None, existing_map: GameMap = None
    ) -> None:
        self.env_cfg = env_cfg
        self.height = self.env_cfg.map_size
        self.width = self.env_cfg.map_size
        self.seed = seed
        rng = np.random.RandomState(seed=seed)
        if existing_map is None:
            self.gen_map(seed, rng, env_cfg)
        else:
            self.map = existing_map
        self.post_map_gen(env_cfg, rng)

    def gen_map(self, seed, rng, env_cfg):

        map_type = rng.choice(["Cave", "Mountain"])
        map_distribution_type = rng.choice(
            [
                "high_ice_high_ore",
                "high_ice_low_ore",
                "low_ice_high_ore",
                "low_ice_low_ore",
            ]
        )
        symmetry = rng.choice(["horizontal", "vertical"])
        self.map = GameMap.random_map(
            seed=seed,
            symmetry=symmetry,
            map_type=map_type,
            map_distribution_type=map_distribution_type,
            width=self.width,
            height=self.height,
        )
        # remove bottom once generator is fully ready
        self.map.rubble = self.map.rubble.astype(int)
        self.map.ore = self.map.ore.astype(int)
        self.map.ice = self.map.ice.astype(int)

    def post_map_gen(self, env_cfg, rng):
        self.factories_per_team = rng.randint(
            env_cfg.MIN_FACTORIES, env_cfg.MAX_FACTORIES + 1
        )

        self.lichen = np.zeros((self.height, self.width), dtype=int)
        # ownership of lichen by factory id, a simple mask
        # -1 = no ownership
        self.lichen_strains = -np.ones((self.height, self.width), dtype=int)
        # self.units_map: np.ndarray = np.zeros((self.height, self.width))
        self.units_map: Dict[str, List[Unit]] = defaultdict(list)
        # TODO consider unit occupancy map, may speed up computation by storing more info in a form indexable in numpy arrays

        # maps center of factory to the factory
        self.factory_map: Dict[str, "Factory"] = dict()
        # != -1 if a factory tile is on the location. Equals factory number id
        self.factory_occupancy_map = -np.ones((self.height, self.width), dtype=int)

        # Valid spawn locations
        self.valid_spawns_mask = np.ones((self.height, self.width), dtype=bool)
        resource_mask = (self.map.ice != 0) | (self.map.ore != 0)
        # generate all 9 shifts of the resource mask. Below is a simple solution to add 1s around any existing 1s in resource_mask
        init_mask = np.zeros((self.height + 2, self.width + 2), dtype=bool)
        for delta in np.array(
            [
                [0, 0],
                [0, 1],
                [0, -1],
                [1, 1],
                [1, 0],
                [1, -1],
                [-1, -1],
                [-1, 0],
                [-1, 1],
            ]
        ):
            s0 = delta[0] + 1
            s1 = delta[1] + 1
            e0 = -1 + delta[0]
            e1 = -1 + delta[1]
            if e0 == 0:
                e0 = None
            if e1 == 0:
                e1 = None
            init_mask[s0:e0, s1:e1] = init_mask[s0:e0, s1:e1] | resource_mask
        resource_mask = init_mask[1:-1, 1:-1]
        self.valid_spawns_mask[resource_mask] = False
        self.valid_spawns_mask[0] = False
        self.valid_spawns_mask[-1] = False
        self.valid_spawns_mask[:, 0] = False
        self.valid_spawns_mask[:, -1] = False
        # grow the 0s by 1 radius

    def pos_hash(self, pos: Position):
        return f"{pos.x},{pos.y}"

    def get_units_at(self, pos: Position):
        pos_hash = self.pos_hash(pos)
        if pos_hash in self.units_map:
            return self.units_map[pos_hash]
        return None

    def get_factory_at(self, state: State, pos: Position):
        f_id = self.factory_occupancy_map[pos.x, pos.y]
        if f_id != -1:
            unit_id = f"factory_{f_id}"
            if unit_id in state.factories["player_0"]:
                return state.factories["player_0"][unit_id]
            else:
                return state.factories["player_1"][unit_id]
        return None

    @property
    def rubble(self) -> np.ndarray:
        return self.map.rubble

    @property
    def ice(self) -> np.ndarray:
        return self.map.ice

    @property
    def ore(self):
        return self.map.ore

    def state_dict(self) -> BoardStateDict:
        return dict(
            rubble=self.rubble.copy(),
            ore=self.ore.copy(),
            ice=self.ice.copy(),
            lichen=self.lichen.copy(),
            lichen_strains=self.lichen_strains.copy(),
            valid_spawns_mask=self.valid_spawns_mask.copy(),
            factories_per_team=self.factories_per_team,
        )



import asyncio
import json
import os
import os.path as osp
import time
from argparse import Namespace
from subprocess import Popen

from luxai_runner.ext_to_command import ext_to_command
from luxai_runner.logger import Logger
from luxai_runner.process import BotProcess


class Bot:
    def __init__(
        self,
        main_file_path: str,
        agent: str,
        agent_idx: str,
        verbose: int = 1,
        direct_import_python_bots=False,
    ) -> None:
        """
        if direct_import_python_bots is True, will directly import the python agents and call their agent_fn functions.
        """
        self.main_file_path = main_file_path
        self.file_ext = osp.splitext(self.main_file_path)[-1]
        if self.file_ext not in ext_to_command:
            raise ValueError(
                f"{self.file_ext} is not a known file extension, we don't know what command to use. \
                Edit https://github.com/Lux-AI-Challenge/Lux-Design-S2/blob/main/luxai_runner/ext_to_command.py \
                    to support a new extension"
            )
        self.command = ext_to_command[self.file_ext]
        self.is_python = self.file_ext == ".py"
        self.agent = agent
        self.agent_idx = agent_idx
        self.verbose = verbose
        self.direct_import_python_bots = direct_import_python_bots
        self.proc = BotProcess(
            self.command,
            self.main_file_path,
            verbose=verbose,
            direct_import_python_bots=direct_import_python_bots,
        )
        # timing
        self.remainingOverageTime = 60
        self.time_per_step = 3

        self.log = Logger(
            identifier=f"{self.agent}, {self.main_file_path}", verbosity=verbose
        )

    async def step(self, obs, step: int, reward: float = 0, info=dict()):
        stime = time.time()
        import copy

        observations = copy.deepcopy(
            dict(
                obs=obs,
                step=step,
                remainingOverageTime=self.remainingOverageTime,
                player=self.agent,
                reward=float(reward),
                info=info,
            )
        )
        stderr = None
        try:
            if self.direct_import_python_bots and self.command == "python":
                env_cfg = None
                if "env_cfg" in info:
                    env_cfg = observations["info"]["env_cfg"]
                observations = Namespace(**observations)
                observations.obs = json.dumps(observations.obs)
                action = self.proc.agent_fn(observations, dict(env_cfg=env_cfg))
            else:
                data = json.dumps(observations)
                action, stderr = await asyncio.wait_for(
                    self.proc.write(f"{data}\n"),
                    timeout=self.remainingOverageTime + self.time_per_step,
                )
        except asyncio.TimeoutError:
            action, stderr = None, None
        time_used = time.time() - stime

        if stderr != "" and stderr is not None:
            self.log.err(f"stderr:\n{stderr}")

        over_time = time_used - self.time_per_step
        if over_time > 0:
            self.remainingOverageTime -= over_time

        if self.remainingOverageTime <= 0 or action is None:
            self.log.err("timed out...")
            action = None
        else:
            try:
                if isinstance(action, dict):
                    return action
                action = json.loads(action)
                if not isinstance(action, dict):
                    raise ValueError("")
            except:
                self.log.err(f"cannot parse action '{action}'")
                action = None
        return action



import asyncio
import atexit
import os
import sys
from datetime import datetime
from queue import Empty, Queue
from subprocess import PIPE, STDOUT, Popen
from threading import Thread
from typing import IO

from luxai_runner.logger import Logger


class BotProcess:
    def __init__(
        self,
        command: str,
        file_path: str,
        verbose: int = 2,
        live_log: str = True,
        direct_import_python_bots=False,
    ) -> None:
        self.command = command
        if self.command == "./":
            self.is_binary = True
        else:
            self.is_binary = False
        self.file_path = file_path
        self.log = Logger(identifier="", verbosity=verbose)
        self.live_log = live_log
        self.direct_import_python_bots = direct_import_python_bots

        if self.direct_import_python_bots and self.command == "python":
            import importlib.util
            import sys

            sys.path.append(os.path.dirname(file_path))
            spec = importlib.util.spec_from_file_location("bot", file_path)
            foo = importlib.util.module_from_spec(spec)
            sys.modules["bot"] = foo
            spec.loader.exec_module(foo)
            # only use this locally with trusted agents!
            self.agent_fn = foo.agent_fn

    async def start(self):
        if self.direct_import_python_bots and self.command == "python":
            return
        cwd = os.path.dirname(self.file_path)
        if cwd == "":
            cwd = "."
        self.log.info(f"Beginning {self.command} {self.file_path}")
        # self._agent_process = Popen([self.command, os.path.basename(self.file_path)], stdin=PIPE, stdout=PIPE, stderr=PIPE, cwd=cwd)

        base_file_path = os.path.basename(self.file_path)
        if self.is_binary:
            self._agent_process = await asyncio.create_subprocess_exec(
                f"./{base_file_path}",
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                cwd=cwd,
                limit=1024 * 128,
            )
        else:
            if self.command == "java" and base_file_path.endswith(".java"):
                base_file_path = base_file_path[:-5]
            self._agent_process = await asyncio.create_subprocess_exec(
                self.command,
                base_file_path,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                cwd=cwd,
                limit=1024 * 128,
            )
        self.log.info(f"Started {self.command} {self.file_path}")
        # following 4 lines from https://stackoverflow.com/questions/375427/a-non-blocking-read-on-a-subprocess-pipe-in-python
        # used to track stderr data asynchronously
        self._q = Queue()

        def enqueue_output(out: IO, queue: Queue):
            line = out.readline()
            self._q.put(line)

        # self._t = Thread(target=enqueue_output, args=(self._agent_process.stderr, self._q))
        # self._t.daemon = True  # thread dies with the program
        # self._t.start()
        self.stderr_queue = []

        async def log_stream(stream):
            while not stream.at_eof():
                data = await stream.readline()
                line = data.decode()

                if stream.at_eof() and len(line) == 0:
                    break
                elif self.live_log:
                    self.log.err(line, end="")
                else:
                    self.stderr_queue.append(line)

        asyncio.create_task(log_stream(self._agent_process.stderr))
        # await asyncio.gather(watch(self._agent_process.stderr, 'E:'))

    async def write(self, msg: str):
        self._agent_process.stdin.write(msg.encode())
        stdout, stderr = await asyncio.gather(
            self._agent_process.stdout.readline(), self.stderr()
        )
        return stdout.decode(), stderr

    async def receive(self) -> str:
        res = (await self._agent_process.stdout.readline()).decode()
        return res

    async def stderr(self):
        r = "".join(self.stderr_queue)
        self.stderr_queue.clear()
        return r
        # while True:
        #     try:
        #         line = self._q.get_nowait()
        #     except Empty:
        #         # no standard error received, break
        #         break
        #     else:
        #         # standard error output received, print it out
        #         stderrs.append((await line).decode())
        #         # stderrs.append(line.decode())
        # return " ".join(stderrs)

    async def cleanup(self):
        self._agent_process._transport.close()
        self._agent_process.terminate()
        await self._agent_process.wait()



from dataclasses import dataclass, field
from typing import Dict, List


@dataclass
class TournamentConfig:
    name: str = "luxai_s2_tourney"
    tournament_type: str = "leadeboard"  # roundrobin, leaderboard
    ranking_system: str = "elo"  # elo, win/loss, TODO: Bradley-Terry
    ranking_system_cfg: Dict = field(default_factory=dict)
    matchmaking_system: str = "random"

    agents: List[str] = field(default_factory=list)

    agents_per_episode: List[int] = field(default_factory=lambda: [2])

    max_concurrent_episodes: int = 4



import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="luxai-s2",
    author="Lux AI Challenge",
    description="The Lux AI Challenge Season 2",
    license="MIT",
    keywords="reinforcement-learning machine-learning ai",
    url="https://github.com/Lux-AI-Challenge/Lux-Design-S2",
    long_description="Code for the Lux AI Challenge Season 2",
    packages=find_packages(exclude="kits"),
    entry_points={"console_scripts": ["luxai-s2 = luxai_runner.cli:main"]},
    version="2.0.5",
    python_requires=">=3.7",
    install_requires=[
        "numpy",
        "pygame",
        "termcolor",
        "matplotlib",
        "pettingzoo",
        "vec_noise",
        "omegaconf",
        "gym==0.21.0",
        "scipy",
        "importlib-metadata<5.0" # fixes bug where they deprecated an endpoint that openai gym uses
    ],
)



from typing import Callable, Dict

import gym
import numpy as np
import numpy.typing as npt
from gym import spaces

import luxai_s2.env
from luxai_s2.env import LuxAI_S2
from luxai_s2.state import ObservationStateDict
from luxai_s2.unit import ActionType, BidActionType, FactoryPlacementActionType
from luxai_s2.utils import my_turn_to_place_factory
from luxai_s2.wrappers.controllers import (
    Controller,
    SimpleDiscreteController,
    SimpleSingleUnitDiscreteController,
)


class SingleUnitObservationWrapper(gym.ObservationWrapper):
    """
    A state based observation to work with in pair with the SimpleSingleUnitDiscreteController

    It contains info only on the first heavy unit, the first factory you own, and some useful features. If there are no owned heavy units the observation is just zero.
    No information about the opponent is included


    Included features:
    - First heavy unit's stats
    - Position of closest ice tile
    - First factory

    """

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = spaces.Box(-999, 999, shape=(13,))

    def observation(
        self, obs: Dict[str, ObservationStateDict]
    ) -> Dict[str, npt.NDArray]:
        observation = dict()
        shared_obs = obs["player_0"]
        ice_map = shared_obs["board"]["ice"]
        ice_tile_locations = np.argwhere(ice_map == 1)
        for agent in obs.keys():
            factories = shared_obs["factories"][agent]
            factory_vec = np.zeros(2)
            for k in factories.keys():
                factory = factories[k]
                factory_vec = np.array(factory["pos"]) / self.env.state.env_cfg.map_size
                break
            units = shared_obs["units"][agent]
            for k in units.keys():
                unit = units[k]
                cargo_space = self.env.state.env_cfg.ROBOTS[
                    unit["unit_type"]
                ].CARGO_SPACE
                battery_cap = self.env.state.env_cfg.ROBOTS[
                    unit["unit_type"]
                ].BATTERY_CAPACITY
                cargo_vec = np.array(
                    [
                        unit["power"] / battery_cap,
                        unit["cargo"]["ice"] / cargo_space,
                        unit["cargo"]["ore"] / cargo_space,
                        unit["cargo"]["water"] / cargo_space,
                        unit["cargo"]["metal"] / cargo_space,
                    ]
                )
                unit_type = (
                    0 if unit["unit_type"] == "LIGHT" else 1
                )  # note that build actions use 0 to encode Light

                pos = np.array(unit["pos"]) / self.env.state.env_cfg.map_size
                unit_vec = np.concatenate(
                    [pos, [unit_type], cargo_vec, [unit["team_id"]]], axis=-1
                )
                # engineered features

                # compute closest ice tile
                ice_tile_distances = np.mean(
                    (ice_tile_locations - np.array(unit["pos"])) ** 2, 1
                )
                closest_ice_tile = (
                    ice_tile_locations[np.argmin(ice_tile_distances)]
                    / self.env.state.env_cfg.map_size
                )
                obs_vec = np.concatenate(
                    [unit_vec, factory_vec - pos, closest_ice_tile - pos], axis=-1
                )

                observation[agent] = obs_vec
                break
            if agent not in observation:
                observation[agent] = np.zeros(13)
        return observation


class ImageObservationWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        obs_dims = 23  # see _convert_obs function for how this is computed
        self.map_size = self.env.env_cfg.map_size
        self.observation_space = spaces.Box(
            -999, 999, shape=(self.map_size, self.map_size, obs_dims)
        )

    def observation(
        self, obs: Dict[str, ObservationStateDict]
    ) -> Dict[str, npt.NDArray]:
        shared_obs = obs["player_0"]
        unit_mask = np.zeros((self.map_size, self.map_size, 1))
        unit_data = np.zeros(
            (self.map_size, self.map_size, 9)
        )  # power(1) + cargo(4) + unit_type(1) + unit_pos(2) + team(1)
        factory_mask = np.zeros_like(unit_mask)
        factory_data = np.zeros(
            (self.map_size, self.map_size, 8)
        )  # power(1) + cargo(4) + factory_pos(2) + team(1)
        for agent in obs.keys():
            factories = shared_obs["factories"][agent]
            units = shared_obs["units"][agent]

            for unit_id in units.keys():
                unit = units[unit_id]
                # we encode everything but unit_id or action queue
                cargo_space = self.env.state.env_cfg.ROBOTS[
                    unit["unit_type"]
                ].CARGO_SPACE
                battery_cap = self.env.state.env_cfg.ROBOTS[
                    unit["unit_type"]
                ].BATTERY_CAPACITY
                cargo_vec = np.array(
                    [
                        unit["power"] / battery_cap,
                        unit["cargo"]["ice"] / cargo_space,
                        unit["cargo"]["ore"] / cargo_space,
                        unit["cargo"]["water"] / cargo_space,
                        unit["cargo"]["metal"] / cargo_space,
                    ]
                )
                unit_type = (
                    0 if unit["unit_type"] == "LIGHT" else 1
                )  # note that build actions use 0 to encode Light
                unit_vec = np.concatenate(
                    [unit["pos"], [unit_type], cargo_vec, [unit["team_id"]]], axis=-1
                )
                unit_vec[:2] /= self.env.state.env_cfg.map_size

                # note that all data is stored as map[x, y] format
                unit_data[unit["pos"][0], unit["pos"][1]] = unit_vec
                unit_mask[unit["pos"][0], unit["pos"][1]] = 1

            for unit_id in factories.keys():
                factory = factories[unit_id]
                # we encode everything but strain_id or unit_id
                cargo_vec = np.array(
                    [
                        factory["power"],
                        factory["cargo"]["ice"],
                        factory["cargo"]["ore"],
                        factory["cargo"]["water"],
                        factory["cargo"]["metal"],
                    ]
                )
                cargo_vec = cargo_vec * 1 / 1000

                factory_vec = np.concatenate(
                    [factory["pos"], cargo_vec, [factory["team_id"]]], axis=-1
                )
                factory_vec[:2] /= self.env.state.env_cfg.map_size
                factory_data[factory["pos"][0], factory["pos"][1]] = factory_vec
                factory_mask[factory["pos"][0], factory["pos"][1]] = 1

            image_features = np.concatenate(
                [
                    np.expand_dims(shared_obs["board"]["lichen"], -1)
                    / self.env.state.env_cfg.MAX_LICHEN_PER_TILE,
                    np.expand_dims(shared_obs["board"]["rubble"], -1)
                    / self.env.state.env_cfg.MAX_RUBBLE,
                    np.expand_dims(shared_obs["board"]["ice"], -1),
                    np.expand_dims(shared_obs["board"]["ore"], -1),
                    unit_mask,
                    unit_data,
                    factory_mask,
                    factory_data,
                ],
                axis=-1,
            )

        new_obs = dict()
        for agent in self.all_agents:
            new_obs[agent] = image_features
        return new_obs



from typing import List

from luxai_s2.config import EnvConfig
from luxai_s2.unit import Unit


def is_day(config: EnvConfig, env_step):
    return env_step % config.CYCLE_LENGTH < config.DAY_LENGTH


def get_top_two_power_units(units: List[Unit]):
    most_power_unit: Unit = units[0]
    most_power = -1
    next_most_power_unit: Unit = units[1]
    next_most_power = -1
    for u in units:
        if u.power > most_power:
            next_most_power_unit = most_power_unit
            most_power_unit = u
            most_power = u.power
        elif (
            u.power >= next_most_power
        ):  # >= check since we want to top 2 power units which can tie
            next_most_power_unit = u
            next_most_power = u.power

    return (most_power_unit, next_most_power_unit)


def my_turn_to_place_factory(place_first: bool, step: int):
    if place_first:
        if step % 2 == 1:
            return True
    else:
        if step % 2 == 0:
            return True
    return False



from abc import ABC, abstractclassmethod
from dataclasses import dataclass

import numpy as np


class Rank:
    rating: int


class RankingSystem(ABC):
    def __init__(self) -> None:
        pass

    def init_rank_state(self) -> Rank:
        pass

    def update(
        self, rank_1: Rank, rank_2: Rank, rank_1_score: float, rank_2_score: float
    ):
        pass

    def _rank_headers(self) -> str:
        pass

    def _rank_info(self, rank: Rank) -> str:
        pass


@dataclass
class ELORank(Rank):
    rating: int
    episodes: int


class ELO(RankingSystem):
    def __init__(self, K=32, init_rating=1000, col_length=10) -> None:
        super().__init__()
        self.col_length = col_length
        self.K = K
        self.init_rating = 1000

    def init_rank_state(self) -> Rank:
        return ELORank(rating=self.init_rating, episodes=0)

    def update(self, rank_1: ELORank, rank_2: ELORank, rank_1_score, rank_2_score):
        # Only implements win/loss/ties
        if rank_1_score > rank_2_score:
            rank_1.rating = rank_1.rating + self.K * (
                1 - self._expected_score(rank_1, rank_2)
            )
            rank_2.rating = rank_2.rating + self.K * (
                0 - self._expected_score(rank_2, rank_1)
            )
        elif rank_1_score < rank_2_score:
            rank_1.rating = rank_1.rating + self.K * (
                0 - self._expected_score(rank_1, rank_2)
            )
            rank_2.rating = rank_2.rating + self.K * (
                1 - self._expected_score(rank_2, rank_1)
            )
        else:
            rank_1.rating = rank_1.rating + self.K * (
                0.5 - self._expected_score(rank_1, rank_2)
            )
            rank_2.rating = rank_2.rating + self.K * (
                0.5 - self._expected_score(rank_2, rank_1)
            )
        rank_1.episodes += 1
        rank_2.episodes += 1

    def _expected_score(self, rank_1: ELORank, rank_2: ELORank):
        """
        Returns E_1 = expected score for player 1
        """
        return 1 / (1 + np.power(10, (rank_2.rating - rank_1.rating) / 400))

    def _rank_headers(self) -> str:
        return f"{'Rating':{self.col_length}.{self.col_length}}"

    def _rank_info(self, rank: ELORank) -> str:
        return f"{str(rank.rating):{self.col_length}.{self.col_length}}"


@dataclass
class WinLossRank(Rank):
    rating: int
    wins: int
    ties: int
    losses: int
    episodes: int


class WinLoss(RankingSystem):
    def __init__(
        self, win_points=3, tie_points=1, loss_points=0, col_length=10
    ) -> None:
        super().__init__()
        self.col_length = col_length
        self.win_points = win_points
        self.tie_points = tie_points
        self.loss_points = loss_points

    def init_rank_state(self) -> Rank:
        return WinLossRank(rating=0, episodes=0, wins=0, ties=0, losses=0)

    def update(
        self, rank_1: WinLossRank, rank_2: WinLossRank, rank_1_score, rank_2_score
    ):
        # Only implements win/loss/ties
        if rank_1_score == rank_2_score:
            winner = None
            loser = None
            rank_1.ties += 1
            rank_2.ties += 1
            rank_1.rating = rank_1.rating + self.tie_points
            rank_2.rating = rank_2.rating + self.tie_points
        else:
            winner = rank_1
            loser = rank_2
            if rank_1_score < rank_2_score:
                winner = rank_2
                loser = rank_1
            # not a tie
            winner.wins += 1
            loser.losses += 1
            winner.rating = winner.rating + self.win_points
            loser.rating = loser.rating + self.loss_points

        rank_1.episodes += 1
        rank_2.episodes += 1

    def _rank_headers(self) -> str:
        return f"{'Score':{self.col_length}} | {'Wins':{self.col_length}} | {'Ties':{self.col_length}} | {'Losses':{self.col_length}}"

    def _rank_info(self, rank: WinLossRank) -> str:
        return f"{str(rank.rating):{self.col_length}.{self.col_length}} | {str(rank.wins):{self.col_length}.{self.col_length}} | {str(rank.ties):{self.col_length}.{self.col_length}} | {str(rank.losses):{self.col_length}.{self.col_length}}"
